"""A building block for other db tests."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging

from test_db_base import TestDBBase


logger = logging.getLogger(__name__)


class TestSummaryDBBase(TestDBBase):

    def test_db_cols_have_values(self):
        logger.info("Checking DB %s months table has values", self.db.db_name)
        months_table = self.table_dict['months_table']
        self.assertGreater(months_table.get_col_max(self.db, months_table.hr_avg), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.hr_min), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.hr_max), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.rhr_avg), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.rhr_min), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.rhr_max), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.weight_avg), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.weight_min), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.weight_max), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.stress_avg), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.calories_avg), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.calories_bmr_avg), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.calories_active_avg), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.activities), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.activities_calories), 0)
        self.assertGreater(months_table.get_col_max(self.db, months_table.activities_distance), 0)



"""Object for implementing summary databse objects."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import datetime
from sqlalchemy import Column, Float, Time, Integer, func
from sqlalchemy.ext.hybrid import hybrid_property

import fitfile.conversions as conversions
from idbutils import DbObject


class SummaryBase(DbObject):
    """Base class for implementing summary database objects."""

    view_version = 10
    _col_units = {'hr_avg': 'bpm', 'hr_min': 'bpm', 'hr_max': 'bpm', 'rhr_avg': 'bpm', 'rhr_min': 'bpm', 'rhr_max': 'bpm', 'rr_waking_avg': 'brpm', 'rr_max': 'brpm',
                  'rr_min': 'brpm'}

    hr_avg = Column(Float)
    hr_min = Column(Float)
    hr_max = Column(Float)
    rhr_avg = Column(Float)
    rhr_min = Column(Float)
    rhr_max = Column(Float)
    inactive_hr_avg = Column(Float)
    inactive_hr_min = Column(Float)
    inactive_hr_max = Column(Float)
    weight_avg = Column(Float)
    weight_min = Column(Float)
    weight_max = Column(Float)
    intensity_time = Column(Time, nullable=False, default=datetime.time.min)
    moderate_activity_time = Column(Time, nullable=False, default=datetime.time.min)
    vigorous_activity_time = Column(Time, nullable=False, default=datetime.time.min)
    intensity_time_goal = Column(Time, nullable=False, default=datetime.time.min)
    steps = Column(Integer)
    steps_goal = Column(Integer)
    floors = Column(Float)
    floors_goal = Column(Float)
    sleep_avg = Column(Time, nullable=False, default=datetime.time.min)
    sleep_min = Column(Time, nullable=False, default=datetime.time.min)
    sleep_max = Column(Time, nullable=False, default=datetime.time.min)
    rem_sleep_avg = Column(Time, nullable=False, default=datetime.time.min)
    rem_sleep_min = Column(Time, nullable=False, default=datetime.time.min)
    rem_sleep_max = Column(Time, nullable=False, default=datetime.time.min)
    stress_avg = Column(Integer)
    calories_avg = Column(Integer)
    calories_bmr_avg = Column(Integer)
    calories_active_avg = Column(Integer)
    calories_goal = Column(Integer)
    calories_active_avg = Column(Integer)
    calories_consumed_avg = Column(Integer)
    activities = Column(Integer)
    activities_calories = Column(Integer)
    activities_distance = Column(Integer)
    hydration_goal = Column(Integer)
    hydration_avg = Column(Integer)
    hydration_intake = Column(Integer)
    sweat_loss_avg = Column(Integer)
    sweat_loss = Column(Integer)
    spo2_avg = Column(Float)
    spo2_min = Column(Float)
    rr_waking_avg = Column(Float)
    rr_max = Column(Float)
    rr_min = Column(Float)

    @hybrid_property
    def intensity_time_mins(self):
        """Return intensity time as minutes."""
        return (conversions.time_to_secs(self.intensity_time) / 60) if self.intensity_time is not None else 0

    @intensity_time_mins.expression
    def intensity_time_mins(cls):
        """Return intensity time as minutes."""
        return (cls._secs_from_time(cls.intensity_time) / 60)

    @hybrid_property
    def intensity_time_goal_mins(self):
        """Return intensity time as minutes."""
        return (conversions.time_to_secs(self.intensity_time_goal) / 60) if self.intensity_time_goal is not None else 0

    @intensity_time_goal_mins.expression
    def intensity_time_goal_mins(cls):
        """Return intensity time as minutes."""
        return (cls._secs_from_time(cls.intensity_time_goal) / 60)

    @hybrid_property
    def intensity_time_goal_percent(self):
        """Return the percentage of intensity time goal achieved."""
        if self.intensity_time and self.intensity_time_goal:
            intensity_time = conversions.time_to_secs(self.intensity_time)
            intensity_time_goal = conversions.time_to_secs(self.intensity_time_goal)
            if intensity_time and intensity_time_goal:
                return (intensity_time * 100) / intensity_time_goal
        return 0.0

    @intensity_time_goal_percent.expression
    def intensity_time_goal_percent(cls):
        """Return the percentage of intensity time goal achieved."""
        return func.round((cls._secs_from_time(cls.intensity_time) * 100) / cls._secs_from_time(cls.intensity_time_goal))

    @hybrid_property
    def steps_goal_percent(self):
        """Return the percentage of steps goal achieved."""
        return (self.steps * 100) / self.steps_goal if self.steps and self.steps_goal else 0.0

    @steps_goal_percent.expression
    def steps_goal_percent(cls):
        """Return the percentage of steps goal achieved."""
        return func.round((cls.steps * 100) / cls.steps_goal)

    @hybrid_property
    def floors_goal_percent(self):
        """Return the percentage of floors goal achieved."""
        return (self.floors * 100) / self.floors_goal if self.floors and self.floors_goal else 0.0

    @floors_goal_percent.expression
    def floors_goal_percent(cls):
        """Return the percentage of floors goal achieved."""
        return func.round((cls.floors * 100) / cls.floors_goal)

    @classmethod
    def create_summary_view(cls, db, selectable):
        """Create a view in the database from the passed in selectable."""
        cls._create_view_from_selectable(db, cls._get_default_view_name(), selectable, cls.time_col.desc())

    @classmethod
    def __create_weeks_months_years_selectable(cls, days_count):
        selectable = [
            cls.time_col.label('first_day'),
            cls.round_col('rhr_avg', 'rhr'),
            cls.round_col('inactive_hr_avg', 'inactive_hr'),
            cls.round_col('weight_avg', 'weight'),
            cls.intensity_time.label('intensity_time'),
            cls.moderate_activity_time.label('moderate_activity_time'),
            cls.vigorous_activity_time.label('vigorous_activity_time'),
            cls.steps.label('total_steps'),
            cls.round_col_txt(f'round(steps / {days_count})', 'steps_avg'),
            cls.round_col_txt('round((steps * 100) / steps_goal)', 'steps_goal_percent'),
            cls.round_col_txt('floors', 'total_floors'),
            cls.round_col_txt(f'round(floors / {days_count})', 'floors_avg'),
            cls.round_col_txt('round((floors * 100) / floors_goal)', 'floors_goal_percent'),
            cls.sleep_avg.label('sleep_avg'),
            cls.rem_sleep_avg.label('rem_sleep_avg'),
            cls.round_col('stress_avg'),
            cls.round_col('calories_avg'),
            cls.round_col('calories_bmr_avg'),
            cls.round_col('calories_active_avg'),
            cls.round_col('calories_consumed_avg'),
            cls.round_col('calories_goal'),
            cls.activities.label('activities'),
            cls.round_col('activities_calories'),
            cls.round_col('activities_distance'),
            cls.round_col('hydration_goal'),
            cls.round_col('hydration_avg'),
            cls.round_col('sweat_loss_avg'),
            cls.round_col('sweat_loss_avg'),
            cls.round_col('spo2_avg'),
            cls.round_col('rr_waking_avg'),
        ]
        return selectable

    @classmethod
    def create_years_view(cls, db):
        """Create a yearly summary view in the database."""
        cls.create_summary_view(db, cls.__create_weeks_months_years_selectable(days_count=365))

    @classmethod
    def create_months_view(cls, db):
        """Create a monthly summary view in the database."""
        cls.create_summary_view(db, cls.__create_weeks_months_years_selectable(days_count=28))

    @classmethod
    def create_weeks_view(cls, db):
        """Create a weekly summary view in the database."""
        cls.create_summary_view(db, cls.__create_weeks_months_years_selectable(days_count=7))

    @classmethod
    def create_days_view(cls, db):
        """Create a daily summary view in the database."""
        cols = [
            cls.time_col.label('day'),
            cls.round_col('hr_avg'),
            cls.round_col('hr_min'),
            cls.round_col('hr_max'),
            cls.round_col('rhr_avg', 'rhr'),
            cls.round_col('inactive_hr_avg', 'inactive_hr'),
            cls.round_col('weight_avg', 'weight'),
            cls.intensity_time.label('intensity_time'), cls.moderate_activity_time.label('moderate_activity_time'), cls.vigorous_activity_time.label('vigorous_activity_time'),
            cls.steps.label('steps'),
            # cls.steps_goal_percent,
            cls.round_col_txt('round((steps * 100) / steps_goal)', 'steps_goal_percent'),
            cls.round_col('floors'),
            # cls.floors_goal_percent,
            cls.round_col_txt('round((floors * 100) / floors_goal)', 'floors_goal_percent'),
            cls.sleep_avg.label('sleep_avg'),
            cls.rem_sleep_avg.label('rem_sleep_avg'),
            cls.round_col('stress_avg'),
            cls.round_col('calories_avg'),
            cls.round_col('calories_bmr_avg'),
            cls.round_col('calories_active_avg'),
            cls.round_col('calories_consumed_avg'),
            cls.round_col('calories_goal'),
            cls.activities.label('activities'),
            cls.round_col('activities_calories'),
            cls.round_col('activities_distance'),
            cls.round_col('hydration_goal'),
            cls.round_col('hydration_avg'),
            cls.round_col('sweat_loss_avg'),
            cls.round_col('spo2_avg'),
            cls.round_col('spo2_min'),
            cls.round_col('rr_waking_avg'),
            cls.round_col('rr_max'),
            cls.round_col('rr_min')
        ]
        cls.create_summary_view(db, cols)



"""Class that takes a parsed monitoring FIT file object and imports it into a database."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
import sys
import traceback
import datetime

import fitfile
import idbutils

from .garmindb import File
from .garmindb import MonitoringDb, Monitoring, MonitoringInfo, MonitoringHeartRate, MonitoringIntensity, MonitoringClimb, MonitoringRespirationRate, MonitoringPulseOx
from .fit_file_processor import FitFileProcessor


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class MonitoringFitFileProcessor(FitFileProcessor):
    """Class that takes a parsed monitoring FIT file object and imports it into a database."""

    def __init__(self, db_params, plugin_manager, debug=0):
        """
        Return a new FitFileProcessor instance.

        Paramters:
        db_params (dict): database access configuration
        debug (Boolean): if True, debug logging is enabled
        """
        root_logger.info("Debug: %s", debug)
        super().__init__(db_params, plugin_manager, debug)
        self.garmin_mon_db = MonitoringDb(db_params, self.debug - 1)

    def write_file(self, fit_file):
        """Given a Fit File object, write all of its messages to the DB."""
        with self.garmin_db.managed_session() as self.garmin_db_session, self.garmin_mon_db.managed_session() as self.garmin_mon_db_session:
            self._write_message_types(fit_file, fit_file.message_types)

    def _write_monitoring_info_entry(self, fit_file, message_fields):
        activity_types = message_fields.activity_type
        if isinstance(activity_types, list):
            for index, activity_type in enumerate(activity_types):
                entry = {
                    'file_id'                   : File.s_get_id(self.garmin_db_session, fit_file.filename),
                    'timestamp'                 : message_fields.local_timestamp,
                    'activity_type'             : activity_type,
                    'resting_metabolic_rate'    : message_fields.get('resting_metabolic_rate'),
                    'cycles_to_distance'        : message_fields.cycles_to_distance[index],
                    'cycles_to_calories'        : message_fields.cycles_to_calories[index]
                }
                MonitoringInfo.s_insert_or_update(self.garmin_mon_db_session, entry)

    def _write_monitoring_entry(self, fit_file, message_fields):
        # Only include not None values so that we match and update only if a table's columns if it has values.
        entry = idbutils.list_and_dict.dict_filter_none_values(message_fields)
        timestamp = fit_file.utc_datetime_to_local(message_fields.timestamp)
        # Hack: daily monitoring summaries appear at 00:00:00 localtime for the PREVIOUS day. Subtract a second so they appear in the previous day.
        if timestamp.time() == datetime.time.min:
            timestamp = timestamp - datetime.timedelta(seconds=1)
        entry['timestamp'] = timestamp
        logger.debug("monitoring entry: %r", entry)
        try:
            intersection = MonitoringHeartRate.intersection(entry)
            if len(intersection) > 1 and intersection['heart_rate'] > 0:
                MonitoringHeartRate.s_insert_or_update(self.garmin_mon_db_session, intersection)
            intersection = MonitoringIntensity.intersection(entry)
            if len(intersection) > 1:
                MonitoringIntensity.s_insert_or_update(self.garmin_mon_db_session, intersection)
            intersection = MonitoringClimb.intersection(entry)
            if len(intersection) > 1:
                MonitoringClimb.s_insert_or_update(self.garmin_mon_db_session, intersection)
            intersection = Monitoring.intersection(entry)
            if len(intersection) > 1:
                Monitoring.s_insert_or_update(self.garmin_mon_db_session, intersection)
        except ValueError:
            logger.error("write_monitoring_entry: ValueError for %r: %s", entry, traceback.format_exc())
        except Exception:
            logger.error("Exception on monitoring entry: %r: %s", entry, traceback.format_exc())

    def _write_respiration_entry(self, fit_file, message_fields):
        logger.debug("respiration message: %r", message_fields)
        rr = message_fields.get('respiration_rate')
        if rr > 0:
            respiration = {
                'timestamp' : fit_file.utc_datetime_to_local(message_fields.timestamp),
                'rr'        : rr,
            }
            if fit_file.type is fitfile.FileType.monitoring_b:
                MonitoringRespirationRate.s_insert_or_update(self.garmin_mon_db_session, respiration)
            else:
                raise(ValueError(f'Unexpected file type {repr(fit_file.type)} for respiration message'))

    def _write_pulse_ox_entry(self, fit_file, message_fields):
        logger.debug("pulse_ox message: %r", message_fields)
        if fit_file.type is fitfile.FileType.monitoring_b:
            pulse_ox = message_fields.get('pulse_ox')
            if pulse_ox is not None:
                pulse_ox_entry = {
                    'timestamp': fit_file.utc_datetime_to_local(message_fields.timestamp),
                    'pulse_ox': pulse_ox,
                }
                MonitoringPulseOx.s_insert_or_update(self.garmin_mon_db_session, pulse_ox_entry)
        else:
            raise(ValueError(f'Unexpected file type {repr(fit_file.type)} for pulse ox'))



"""Class for copying data from a USB mounted Garmin device."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os
import sys
import shutil
from tqdm import tqdm
import logging

import fitfile
from idbutils import FileProcessor

from .config_manager import ConfigManager


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))


class Copy(object):
    """Class for copying data from a USB mounted Garmin device."""

    def __init__(self, device_mount_dir):
        """Create a Copy object given the directory where the Garmin USB device is mounted."""
        self.device_mount_dir = device_mount_dir
        if not os.path.exists(self.device_mount_dir):
            raise RuntimeError(f'Device mount directory {self.device_mount_dir} not found')
        if not os.path.isdir(self.device_mount_dir):
            raise RuntimeError(f'Device mount directory {self.device_mount_dir} not a directory')

    def __copy(self, src_dir, dest_dir, latest=False):
        """Copy FIT files from a USB mounted Garmin device to the given directory."""
        file_names = FileProcessor.dir_to_files(src_dir, fitfile.file.name_regex, latest)
        logger.info("Copying files from %s to %s", src_dir, dest_dir)
        for file in tqdm(file_names, unit='files'):
            shutil.copy(file, dest_dir)

    def copy_activities(self, activities_dir, latest=False):
        """Copy activites data FIT files from a USB mounted Garmin device to the given directory."""
        device_activities_dir = ConfigManager.device_activities_dir(self.device_mount_dir)
        self.__copy(device_activities_dir, activities_dir, latest)

    def copy_monitoring(self, monitoring_dir, latest=False):
        """Copy daily monitoring data FIT files from a USB mounted Garmin device to the given directory."""
        device_monitoring_dir = ConfigManager.device_monitoring_dir(self.device_mount_dir)
        self.__copy(device_monitoring_dir, monitoring_dir, latest)

    def copy_sleep(self, monitoring_dir, latest=False):
        """Copy daily sleep data FIT files from a USB mounted Garmin device to the given directory."""
        device_sleep_dir = ConfigManager.device_sleep_dir(self.device_mount_dir)
        self.__copy(device_sleep_dir, monitoring_dir, latest)

    def copy_settings(self, settings_dir):
        """Copy settings FIT files from a USB mounted Garmin device to the given directory."""
        device_settings_dir = ConfigManager.device_settings_dir(self.device_mount_dir)
        self.__copy(device_settings_dir, settings_dir)



"""Enums representing Garmin Connect data types."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import enum
import logging

import fitfile


logger = logging.getLogger(__file__)


class Event(enum.Enum):
    """Garmin Connect event types enum."""

    race            = 1
    recreation      = 2
    special_event   = 3
    training        = 4
    transportation  = 5
    touring         = 6
    geocaching      = 7
    fitness         = 8
    uncategorized   = 9

    @classmethod
    def from_json(cls, json_data):
        """Create a Event enum instance from Garmin Connect JSON data."""
        json_event = json_data['eventType']
        try:
            return cls(json_event['typeId'])
        except ValueError:
            logger.error("Unknown event type: %r", json_event)
            raise


class Sport(enum.Enum):
    """Garmin Connect sport types enum."""

    running                                 = 1
    cycling                                 = 2
    hiking                                  = 3
    other                                   = 4
    mountain_biking                         = 5
    trail_running                           = 6
    street_running                          = 7
    track_running                           = 8
    walking                                 = 9
    road_biking                             = 10
    indoor_cardio                           = 11
    strength_training                       = 13
    casual_walking                          = 15
    speed_walking                           = 16
    top_level                               = 17
    treadmill_running                       = 18
    cyclocross                              = 19
    downhill_biking                         = 20
    track_cycling                           = 21
    recumbent_cycling                       = 22
    indoor_cycling                          = 25
    swimming                                = 26
    lap_swimming                            = 27
    open_water_swimming                     = 28
    fitness_equipment                       = 29
    elliptical                              = 30
    stair_climbing                          = 31
    indoor_rowing                           = 32
    snow_shoe                               = 36
    mountaineering                          = 37
    rowing                                  = 39
    wind_kite_surfing                       = 41
    horseback_riding                        = 44
    driving_general                         = 49
    flying                                  = 52
    paddling                                = 57
    whitewater_rafting_kayaking             = 60
    skating                                 = 62
    inline_skating                          = 63
    resort_skiing_snowboarding              = 67
    backcountry_skiing_snowboarding         = 68
    boating                                 = 75
    sailing                                 = 77
    cross_country_skiing                    = 81
    stand_up_paddleboarding                 = 87
    golf                                    = 88
    bmx                                     = 131
    hunting_fishing                         = 133
    surfing                                 = 137
    wakeboarding                            = 138
    rock_climbing                           = 139
    hang_gliding                            = 140
    tennis                                  = 142
    gravel_cycling                          = 143
    diving                                  = 144
    yoga                                    = 149
    floor_climbing                          = 150
    virtual_ride                            = 152
    virtual_run                             = 153
    obstacle_run                            = 154
    indoor_running                          = 156
    safety                                  = 157
    assistance                              = 158
    incident_detected                       = 159
    ccr_diving                              = 161
    auto_racing                             = 162
    yoga_gym                                = 163
    breathwork                              = 164
    winter_sports                           = 165
    snow_shoe_ws                            = 167
    skating_ws                              = 168
    backcountry_skiing_snowboarding_ws      = 169
    skate_skiing_ws                         = 170
    cross_country_skiing_ws                 = 171
    resort_skiing_snowboarding_ws           = 172

    @classmethod
    def __activity_from_json(cls, json_data):
        return json_data['activityType']

    @classmethod
    def __activity_from_details_json(cls, json_data):
        return json_data['activityTypeDTO']

    @classmethod
    def __sport_from_json(cls, json_activity):
        return json_activity['parentTypeId']

    @classmethod
    def __subsport_from_json(cls, json_activity):
        return json_activity['typeId']

    @classmethod
    def from_json(cls, json_data):
        """Create a Sport enum instance from Garmin Connect JSON data."""
        json_activity = cls.__activity_from_json(json_data)
        try:
            return Sport(cls.__sport_from_json(json_activity))
        except ValueError:
            logger.error("Unknown sport type: %r", json_activity)

    @classmethod
    def from_details_json(cls, json_data):
        """Create a Sport enum instance from Garmin Connect JSON details data."""
        json_activity = cls.__activity_from_details_json(json_data)
        try:
            return Sport(cls.__sport_from_json(json_activity))
        except ValueError:
            logger.error("Unknown sport type: %r", json_activity)

    @classmethod
    def subsport_from_json(cls, json_data):
        """Create a Sport enum instance from Garmin Connect subsport JSON data."""
        json_activity = cls.__activity_from_json(json_data)
        try:
            return Sport(cls.__subsport_from_json(json_activity))
        except ValueError:
            logger.error("Unknown subsport type: %r", json_activity)

    @classmethod
    def subsport_from_details_json(cls, json_data):
        """Create a Sport enum instance from Garmin Connect subsport JSON details data."""
        json_activity = cls.__activity_from_details_json(json_data)
        try:
            return Sport(cls.__subsport_from_json(json_activity))
        except ValueError:
            logger.error("Unknown subsport type: %r", json_activity)


def convert_gc_sport_to_fit(gc_sport, gc_sub_sport):
    """Convert sport and subsport values from Garmin Connect to Fit values."""
    remap_gc_sub_sport_to_fit = {
        Sport.mountain_biking                         : fitfile.SubSport.mountain,
        Sport.trail_running                           : fitfile.SubSport.trail,
        Sport.street_running                          : fitfile.SubSport.street,
        Sport.track_running                           : fitfile.SubSport.track,
        Sport.road_biking                             : fitfile.SubSport.road,
        Sport.casual_walking                          : fitfile.SubSport.casual_walking,
        Sport.speed_walking                           : fitfile.SubSport.speed_walking,
        Sport.treadmill_running                       : fitfile.SubSport.treadmill,
        Sport.downhill_biking                         : fitfile.SubSport.downhill,
        Sport.track_cycling                           : fitfile.SubSport.track_cycling,
        Sport.recumbent_cycling                       : fitfile.SubSport.recumbent,
        Sport.indoor_cycling                          : fitfile.SubSport.indoor_cycling,
        Sport.strength_training                       : fitfile.SubSport.strength_training,
        Sport.cyclocross                              : fitfile.SubSport.cyclocross,
        Sport.lap_swimming                            : fitfile.SubSport.lap_swimming,
        Sport.open_water_swimming                     : fitfile.SubSport.open_water,
        Sport.elliptical                              : fitfile.SubSport.elliptical,
        Sport.stair_climbing                          : fitfile.SubSport.stair_climbing,
        Sport.indoor_rowing                           : fitfile.SubSport.indoor_rowing,
        Sport.bmx                                     : fitfile.SubSport.bmx,
        Sport.gravel_cycling                          : fitfile.SubSport.gravel_cycling,
        Sport.yoga                                    : fitfile.SubSport.yoga,
        Sport.obstacle_run                            : fitfile.SubSport.obstacle,
        Sport.indoor_running                          : fitfile.SubSport.indoor_running,
        Sport.yoga_gym                                : Sport.yoga_gym
    }
    remap_winter_sports = {
        Sport.snow_shoe_ws                            : fitfile.Sport.snowshoeing,
        Sport.skating_ws                              : fitfile.Sport.ice_skating,
        Sport.backcountry_skiing_snowboarding_ws      : Sport.backcountry_skiing_snowboarding,
        Sport.skate_skiing_ws                         : Sport.skate_skiing_ws,
        Sport.cross_country_skiing_ws                 : fitfile.Sport.cross_country_skiing,
        Sport.resort_skiing_snowboarding_ws           : Sport.resort_skiing_snowboarding
    }
    remap_gc_sport_to_fit = {
        Sport.running                                 : fitfile.Sport.running,
        Sport.cycling                                 : fitfile.Sport.cycling,
        Sport.hiking                                  : fitfile.Sport.hiking,
        Sport.walking                                 : fitfile.Sport.walking,
        Sport.indoor_cardio                           : Sport.indoor_cardio,
        Sport.swimming                                : fitfile.Sport.swimming,
        Sport.fitness_equipment                       : fitfile.Sport.fitness_equipment,
        Sport.snow_shoe                               : fitfile.Sport.snowshoeing,
        Sport.mountaineering                          : fitfile.Sport.mountaineering,
        Sport.rowing                                  : fitfile.Sport.rowing,
        Sport.wind_kite_surfing                       : fitfile.Sport.kitesurfing,
        Sport.horseback_riding                        : fitfile.Sport.horseback_riding,
        Sport.driving_general                         : fitfile.Sport.driving,
        Sport.flying                                  : fitfile.Sport.flying,
        Sport.paddling                                : fitfile.Sport.paddling,
        Sport.whitewater_rafting_kayaking             : Sport.whitewater_rafting_kayaking,
        Sport.skating                                 : fitfile.Sport.ice_skating,
        Sport.inline_skating                          : fitfile.Sport.inline_skating,
        Sport.resort_skiing_snowboarding              : Sport.resort_skiing_snowboarding,
        Sport.backcountry_skiing_snowboarding         : Sport.backcountry_skiing_snowboarding,
        Sport.boating                                 : fitfile.Sport.boating,
        Sport.sailing                                 : fitfile.Sport.sailing,
        Sport.cross_country_skiing                    : fitfile.Sport.cross_country_skiing,
        Sport.stand_up_paddleboarding                 : fitfile.Sport.stand_up_paddleboarding,
        Sport.golf                                    : fitfile.Sport.golf,
        Sport.hunting_fishing                         : Sport.hunting_fishing,
        Sport.surfing                                 : fitfile.Sport.surfing,
        Sport.wakeboarding                            : fitfile.Sport.wakeboarding,
        Sport.rock_climbing                           : fitfile.Sport.rock_climbing,
        Sport.hang_gliding                            : fitfile.Sport.hang_gliding,
        Sport.tennis                                  : fitfile.Sport.tennis,
        Sport.diving                                  : Sport.diving,
        Sport.floor_climbing                          : fitfile.Sport.floor_climbing,
        Sport.virtual_ride                            : Sport.virtual_ride,
        Sport.virtual_run                             : Sport.virtual_run,
        Sport.safety                                  : Sport.safety,
        Sport.assistance                              : Sport.assistance,
        Sport.incident_detected                       : Sport.incident_detected,
        Sport.ccr_diving                              : Sport.ccr_diving,
        Sport.auto_racing                             : Sport.auto_racing,
        Sport.breathwork                              : Sport.breathwork,
        Sport.other                                   : Sport.other,
    }
    special_remaps = {
        Sport.other                                   : remap_gc_sport_to_fit.get(gc_sub_sport),
        Sport.top_level                               : remap_gc_sport_to_fit.get(gc_sub_sport),
        Sport.winter_sports                           : remap_winter_sports.get(gc_sub_sport),
    }
    if gc_sport in special_remaps.keys():
        return (special_remaps[gc_sport], fitfile.Sport.generic)
    return (remap_gc_sport_to_fit.get(gc_sport), remap_gc_sub_sport_to_fit.get(gc_sub_sport))


def get_details_sport(json_data):
    """Get the sport and sub-sport and convert to Fit values if possible."""
    sport = Sport.from_details_json(json_data)
    sub_sport = Sport.subsport_from_details_json(json_data)
    return convert_gc_sport_to_fit(sport, sub_sport)


def get_summary_sport(json_data):
    """Get the sport and sub-sport and convert to Fit values if possible."""
    sport = Sport.from_json(json_data)
    sub_sport = Sport.subsport_from_json(json_data)
    return convert_gc_sport_to_fit(sport, sub_sport)



"""Objects for analyzing health data from Garmin devices and genrating summary tables."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import logging
import datetime
import calendar
from tqdm import tqdm

import fitfile
from idbutils.list_and_dict import list_not_none

from garmindb import summarydb
from .garmindb import GarminDb, Attributes, File, Weight, Stress, RestingHeartRate, IntensityHR, Sleep
from .garmindb import MonitoringDb, Monitoring, MonitoringHeartRate, MonitoringIntensity, MonitoringClimb
from .garmindb import ActivitiesDb, Activities, ActivityRecords, ActivityLaps, StepsActivities
from .garmindb import GarminSummaryDb, Summary, DaysSummary, DailySummary, WeeksSummary, MonthsSummary, YearsSummary
from .garmin_connect_config_manager import GarminConnectConfigManager


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
stat_logger = logging.getLogger('stats')
stat_logger.addHandler(logging.FileHandler('stats.txt', 'w'))


class Analyze(object):
    """Object for analyzing health data from Garmin devices."""

    def __init__(self, db_params, debug):
        """Return an instance of the Analyze class."""
        self.garmin_db = GarminDb(db_params, debug)
        self.garmin_mon_db = MonitoringDb(db_params, debug)
        self.garmin_sum_db = GarminSummaryDb(db_params, debug)
        self.sum_db = summarydb.SummaryDb(db_params, debug)
        self.garmin_act_db = ActivitiesDb(db_params, debug)
        self.measurement_system = Attributes.measurements_type(self.garmin_db)
        self.unit_strings = fitfile.units.unit_strings[self.measurement_system]

    def __save_summary_stat(self, name, value):
        Summary.set(self.garmin_sum_db, name, value)
        summarydb.Summary.set(self.sum_db, name, value)

    def __report_file_type(self, file_type):
        records = File.row_count(self.garmin_db, File.type, file_type)
        if records > 0:
            stat_logger.info("%s files: %d", file_type, records)
            self.__save_summary_stat(file_type + '_files', records)

    def __get_files_stats(self):
        records = File.row_count(self.garmin_db)
        stat_logger.info("File records: %d" % records)
        self.__save_summary_stat('files', records)
        for file_type_name in [file_type.name for file_type in File.FileType]:
            self.__report_file_type(file_type_name)

    def __report_sport(self, sport_col, sport):
        fitfile.units.unit_strings[self.measurement_system]
        records = Activities.row_count(self.garmin_act_db, sport_col, sport)
        if records > 0:
            sport_title = sport.title().replace('_', ' ')
            total_distance = Activities.get_col_sum_for_value(self.garmin_act_db, Activities.distance, sport_col, sport)
            if total_distance is None:
                total_distance = 0
                average_distance = 0
            else:
                average_distance = total_distance / records
            distance_units = self.unit_strings[fitfile.units.UnitTypes.distance_long]
            stat_logger.info("%s activities: %d - total distance %d %s average distance %d %s",
                             sport_title, records, total_distance, distance_units, average_distance, distance_units)
            self.__save_summary_stat(sport + '_Activities', records)
            self.__save_summary_stat(sport + '_Miles', total_distance)

    def __get_activities_stats(self):
        stat_logger.info("___Activities Statistics___")
        activities = Activities.row_count(self.garmin_act_db)
        stat_logger.info("Total activities: %d", activities)
        self.__save_summary_stat('Activities', activities)
        laps = ActivityLaps.row_count(self.garmin_act_db)
        stat_logger.info("Activities lap records: %d", laps)
        self.__save_summary_stat('Activity_laps', laps)
        records = ActivityRecords.row_count(self.garmin_act_db)
        stat_logger.info("Activity records: %d", records)
        self.__save_summary_stat('Activity_records', records)
        years = Activities.get_years(self.garmin_act_db)
        stat_logger.info("Years with activities: %d: %s", len(years), years)
        self.__save_summary_stat('Activity_Years', len(years))
        fitness_activities = Activities.row_count(self.garmin_act_db, Activities.type, 'fitness')
        stat_logger.info("Fitness activities: %d", fitness_activities)
        self.__save_summary_stat('Fitness_activities', fitness_activities)
        recreation_activities = Activities.row_count(self.garmin_act_db, Activities.type, 'recreation')
        stat_logger.info("Recreation activities: %d", recreation_activities)
        self.__save_summary_stat('Recreation_activities', recreation_activities)
        sports = list_not_none(Activities.get_col_distinct(self.garmin_act_db, Activities.sport))
        stat_logger.info("Sports: %s", ', '.join(sports))
        sub_sports = list_not_none(Activities.get_col_distinct(self.garmin_act_db, Activities.sub_sport))
        stat_logger.info("SubSports: %s", ', '.join(sub_sports))
        for sport_name in [sport.name for sport in fitfile.Sport]:
            self.__report_sport(Activities.sport, sport_name)

    def __get_col_stats(self, table, col, name, ignore_le_zero=False, time_col=False):
        records = table.row_count(self.garmin_db)
        self.__save_summary_stat('%s_Records' % name, records)
        if time_col:
            maximum = table.get_time_col_max(self.garmin_db, col)
        else:
            maximum = table.get_col_max(self.garmin_db, col)
        self.__save_summary_stat('Max_%s' % name, maximum)
        if time_col:
            minimum = table.get_time_col_min(self.garmin_db, col)
        else:
            minimum = table.get_col_min(self.garmin_db, col, None, None, ignore_le_zero)
        self.__save_summary_stat('Min_%s' % name, minimum)
        if time_col:
            average = table.get_time_col_avg(self.garmin_db, col)
        else:
            average = table.get_col_avg(self.garmin_db, col, None, None, ignore_le_zero)
        self.__save_summary_stat('Avg_%s' % name, average)
        latest = table.get_col_latest(self.garmin_db, col)
        stat_logger.info("%s records: %s max: %s min: %s avg: %s latest: %s", name, records, maximum, minimum, average, latest)

    def __get_monitoring_stats(self):
        stat_logger.info("___Monitoring Statistics___")
        self.__get_col_stats(Weight, Weight.weight, 'Weight')
        self.__get_col_stats(Stress, Stress.stress, 'Stress', True)
        self.__get_col_stats(RestingHeartRate, RestingHeartRate.resting_heart_rate, 'RHR', True)
        self.__get_col_stats(Sleep, Sleep.total_sleep, 'Sleep', True, True)
        self.__get_col_stats(Sleep, Sleep.rem_sleep, 'REM Sleep', True, True)

    def __get_monitoring_days(self, year):
        days = Monitoring.get_days(self.garmin_mon_db, year)
        days_count = len(days)
        if days_count > 0:
            first_day = days[0]
            last_day = days[-1]
            span = last_day - first_day + 1
        else:
            span = 0
        self.__save_summary_stat(str(year) + '_days', days_count)
        self.__save_summary_stat(str(year) + '_days_span', span)
        stat_logger.info("%d Days with data (%d count vs %d span): %s", year, days_count, span, days)
        for index in range(days_count - 1):
            day = int(days[index])
            next_day = int(days[index + 1])
            if next_day != day + 1:
                day_str = str(fitfile.conversions.day_of_the_year_to_datetime(year, day))
                next_day_str = str(fitfile.conversions.day_of_the_year_to_datetime(year, next_day))
                stat_logger.info("Days gap between %d (%s) and %d (%s)", day, day_str, next_day, next_day_str)
        return days_count

    def __get_monitoring_months(self, year):
        months = Monitoring.get_month_names(self.garmin_mon_db, year)
        self.__save_summary_stat(str(year) + '_months', len(months))
        stat_logger.info("%s Months with data (%s): %s", year, len(months), months)

    def __get_monitoring_years(self):
        stat_logger.info("___Monitoring Records Coverage___")
        stat_logger.info("This shows periods that data has been downloaded for. "
                         "Not seeing data for days you know Garmin has data? "
                         "Change the starting day and the number of days in GarminConnectConfig.json and do a full download.")
        years = Monitoring.get_years(self.garmin_mon_db)
        self.__save_summary_stat('Monitoring_Years', len(years))
        stat_logger.info("Monitoring records: %d", Monitoring.row_count(self.garmin_mon_db))
        stat_logger.info("Monitoring Years with data (%d): %s", len(years), years)
        total_days = 0
        for year in years:
            self.__get_monitoring_months(year)
            total_days += self.__get_monitoring_days(year)
        stat_logger.info("Total days with monitoring data: %d", total_days)

    def get_stats(self):
        """Calculate summary statistics."""
        self.__get_files_stats()
        self.__get_activities_stats()
        self.__get_monitoring_stats()
        self.__get_monitoring_years()

    def __populate_hr_intensity(self, day_date, garmin_mon_session, garmin_sum_session, overwrite=False):
        if IntensityHR.s_row_count_for_day(garmin_sum_session, day_date) == 0 or overwrite:
            monitoring_rows = Monitoring._get_for_day(garmin_mon_session, day_date, not_none_col=Monitoring.intensity)
            previous_ts = None
            for monitoring in monitoring_rows:
                # Heart rate value is for one minute, reported at the end of the minute. Only take HR values where the
                # measurement period falls within the activity period.
                if previous_ts is not None and (monitoring.timestamp - previous_ts).total_seconds() > 60:
                    hr_rows = MonitoringHeartRate.s_get_for_period(garmin_mon_session, previous_ts, previous_ts + datetime.timedelta(seconds=60), not_none_col=monitoring.timestamp)
                    for hr in hr_rows:
                        entry = {
                            'timestamp'     : hr.timestamp,
                            'intensity'     : monitoring.intensity,
                            'heart_rate'    : hr.heart_rate
                        }
                        IntensityHR.s_insert_or_update(garmin_sum_session, entry, ignore_none=True)
                previous_ts = monitoring.timestamp

    def __calculate_day_stats(self, day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        stats = DailySummary.get_daily_stats(garmin_session, day_date)
        # prefer getting stats from the daily summary.
        if stats.get('rhr_avg') is None:
            stats.update(RestingHeartRate.get_daily_stats(garmin_session, day_date))
        if stats.get('stress_avg') is None:
            stats.update(Stress.get_daily_stats(garmin_session, day_date))
        if stats.get('intensity_time') is None:
            stats.update(MonitoringIntensity.get_daily_stats(garmin_mon_session, day_date))
        if stats.get('floors') is None:
            stats.update(MonitoringClimb.get_daily_stats(garmin_mon_session, day_date, self.measurement_system))
        if stats.get('steps') is None:
            stats.update(Monitoring.get_daily_stats(garmin_mon_session, day_date))
        stats.update(MonitoringHeartRate.get_daily_stats(garmin_mon_session, day_date))
        stats.update(IntensityHR.get_daily_stats(garmin_sum_session, day_date))
        stats.update(Weight.get_daily_stats(garmin_session, day_date))
        stats.update(Sleep.get_daily_stats(garmin_session, day_date))
        stats.update(Activities.get_daily_stats(garmin_act_session, day_date))
        # save it to the db
        DaysSummary.s_insert_or_update(garmin_sum_session, stats)
        summarydb.DaysSummary.s_insert_or_update(sum_session, stats)

    def __calculate_days(self, year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        days = Monitoring.s_get_days(garmin_mon_session, year)
        for day in tqdm(days, unit='days'):
            day_date = datetime.date(year, 1, 1) + datetime.timedelta(day - 1)
            self.__populate_hr_intensity(day_date, garmin_mon_session, garmin_sum_session)
            self.__calculate_day_stats(day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)

    def __calculate_week_stats(self, day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        stats = DailySummary.get_weekly_stats(garmin_session, day_date)
        # prefer getting stats from the daily summary.
        if stats.get('rhr_avg') is None:
            stats.update(RestingHeartRate.get_weekly_stats(garmin_session, day_date))
        if stats.get('stress_avg') is None:
            stats.update(Stress.get_weekly_stats(garmin_session, day_date))
        if stats.get('intensity_time') is None:
            stats.update(MonitoringIntensity.get_weekly_stats(garmin_mon_session, day_date))
        if stats.get('floors') is None:
            stats.update(MonitoringClimb.get_weekly_stats(garmin_mon_session, day_date, self.measurement_system))
        if stats.get('steps') is None:
            stats.update(Monitoring.get_weekly_stats(garmin_mon_session, day_date))
        stats.update(MonitoringHeartRate.get_weekly_stats(garmin_mon_session, day_date))
        stats.update(IntensityHR.get_weekly_stats(garmin_sum_session, day_date))
        stats.update(Weight.get_weekly_stats(garmin_session, day_date))
        stats.update(Sleep.get_weekly_stats(garmin_session, day_date))
        stats.update(Activities.get_weekly_stats(garmin_act_session, day_date))
        # save it to the db
        WeeksSummary.s_insert_or_update(garmin_sum_session, stats)
        summarydb.WeeksSummary.s_insert_or_update(sum_session, stats)

    def __calculate_weeks(self, year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        for week_starting_day in tqdm(range(1, 365, 7), unit='weeks'):
            day_date = datetime.date(year, 1, 1) + datetime.timedelta(week_starting_day - 1)
            if day_date < datetime.datetime.now().date():
                self.__calculate_week_stats(day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)

    def __calculate_month_stats(self, start_day_date, end_day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        stats = DailySummary.get_monthly_stats(garmin_session, start_day_date, end_day_date)
        # prefer getting stats from the daily summary.
        if 'rhr_avg' in stats:
            stats.update(RestingHeartRate.get_monthly_stats(garmin_session, start_day_date, end_day_date))
        if 'stress_avg' in stats:
            stats.update(Stress.get_monthly_stats(garmin_session, start_day_date, end_day_date))
        if 'intensity_time' in stats:
            stats.update(MonitoringIntensity.get_monthly_stats(garmin_mon_session, start_day_date, end_day_date))
        if 'floors' in stats:
            stats.update(MonitoringClimb.get_monthly_stats(garmin_mon_session, start_day_date, end_day_date, self.measurement_system))
        if 'steps' in stats:
            stats.update(Monitoring.get_monthly_stats(garmin_mon_session, start_day_date, end_day_date))
        stats.update(MonitoringHeartRate.get_monthly_stats(garmin_mon_session, start_day_date, end_day_date))
        stats.update(IntensityHR.get_monthly_stats(garmin_sum_session, start_day_date, end_day_date))
        stats.update(Weight.get_monthly_stats(garmin_session, start_day_date, end_day_date))
        stats.update(Sleep.get_monthly_stats(garmin_session, start_day_date, end_day_date))
        stats.update(Activities.get_monthly_stats(garmin_act_session, start_day_date, end_day_date))
        # save it to the db
        MonthsSummary.s_insert_or_update(garmin_sum_session, stats)
        summarydb.MonthsSummary.s_insert_or_update(sum_session, stats)

    def __calculate_months(self, year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        months = Monitoring.s_get_months(garmin_mon_session, year)
        for month in tqdm(months, unit='months'):
            start_day_date = datetime.date(year, month, 1)
            end_day_date = datetime.date(year, month, calendar.monthrange(year, month)[1])
            self.__calculate_month_stats(start_day_date, end_day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)

    def __calculate_year_stats(self, year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        stats = DailySummary.get_yearly_stats(garmin_session, year)
        # prefer getting stats from the daily summary.
        if 'rhr_avg' in stats:
            stats.update(RestingHeartRate.get_yearly_stats(garmin_session, year))
        if 'stress_avg' in stats:
            stats.update(Stress.get_yearly_stats(garmin_session, year))
        if 'intensity_time' in stats:
            stats.update(MonitoringIntensity.get_yearly_stats(garmin_mon_session, year))
        if 'floors' in stats:
            stats.update(MonitoringClimb.get_yearly_stats(garmin_mon_session, year, self.measurement_system))
        if 'steps' in stats:
            stats.update(Monitoring.get_yearly_stats(garmin_mon_session, year))
        stats.update(MonitoringHeartRate.get_yearly_stats(garmin_mon_session, year))
        stats.update(IntensityHR.get_yearly_stats(garmin_sum_session, year))
        stats.update(Weight.get_yearly_stats(garmin_session, year))
        stats.update(Sleep.get_yearly_stats(garmin_session, year))
        stats.update(Activities.get_yearly_stats(garmin_act_session, year))
        # save it to the db
        YearsSummary.s_insert_or_update(garmin_sum_session, stats)
        summarydb.YearsSummary.s_insert_or_update(sum_session, stats)

    def __calculate_year(self, year):
        with self.garmin_db.managed_session() as garmin_session, self.garmin_mon_db.managed_session() as garmin_mon_session, \
                self.garmin_act_db.managed_session() as garmin_act_session, self.garmin_sum_db.managed_session() as garmin_sum_session, \
                self.sum_db.managed_session() as sum_session:
            # calculate part of the years
            self.__calculate_days(year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)
            self.__calculate_weeks(year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)
            self.__calculate_months(year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)
            # now calculate the year itself
            self.__calculate_year_stats(year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)

    def summary(self):
        """Summarize Garmin health data. Daily, weekly, and monthly, tables will be generated."""
        logger.info("Summary Tables Generation:")
        years = Monitoring.get_years(self.garmin_mon_db)
        for year in years:
            logger.info("Generating table entries for %s", year)
            self.__calculate_year(year)

    def create_dynamic_views(self):
        """Create database views specific to the data in this database."""
        gc_config = GarminConnectConfigManager()
        course_ids = gc_config.course_views('steps')
        if course_ids:
            for course_id in course_ids:
                StepsActivities.create_course_view(self.garmin_act_db, course_id)



"""Class for downloading health data from Garmin Connect."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os
import sys
import re
import logging
import datetime
import time
import tempfile
import zipfile
import json
import cloudscraper
from tqdm import tqdm

import fitfile.conversions as conversions
from idbutils import RestClient, RestException, RestResponseException

from .garmin_connect_config_manager import GarminConnectConfigManager
from .config_manager import ConfigManager


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class Download(object):
    """Class for downloading health data from Garmin Connect."""

    garmin_connect_base_url = "https://connect.garmin.com"
    garmin_connect_enus_url = garmin_connect_base_url + "/en-US"

    garmin_connect_sso_login = 'signin'

    garmin_connect_login_url = garmin_connect_enus_url + "/signin"

    garmin_connect_css_url = 'https://static.garmincdn.com/com.garmin.connect/ui/css/gauth-custom-v1.2-min.css'

    garmin_connect_privacy_url = "//connect.garmin.com/en-U/privacy"

    garmin_connect_user_profile_url = "proxy/userprofile-service/userprofile"
    garmin_connect_wellness_url = "proxy/wellness-service/wellness"
    garmin_connect_sleep_daily_url = garmin_connect_wellness_url + "/dailySleepData"
    garmin_connect_rhr = "proxy/userstats-service/wellness/daily"
    garmin_connect_weight_url = "proxy/weight-service/weight/dateRange"

    garmin_connect_activity_search_url = "proxy/activitylist-service/activities/search/activities"

    garmin_connect_usersummary_url = "proxy/usersummary-service/usersummary"
    garmin_connect_daily_summary_url = garmin_connect_usersummary_url + "/daily"
    garmin_connect_daily_hydration_url = garmin_connect_usersummary_url + "/hydration/allData"

    # https://connect.garmin.com/modern/proxy/usersummary-service/usersummary/hydration/allData/2019-11-29

    garmin_headers = {'NK': 'NT'}

    def __init__(self):
        """Create a new Download class instance."""
        logger.debug("__init__")
        self.session = cloudscraper.CloudScraper()
        self.sso_rest_client = RestClient(self.session, 'sso.garmin.com', 'sso', aditional_headers=self.garmin_headers)
        self.modern_rest_client = RestClient(self.session, 'connect.garmin.com', 'modern', aditional_headers=self.garmin_headers)
        self.activity_service_rest_client = RestClient.inherit(self.modern_rest_client, "proxy/activity-service/activity")
        self.download_service_rest_client = RestClient.inherit(self.modern_rest_client, "proxy/download-service/files")
        self.gc_config = GarminConnectConfigManager()
        self.download_days_overlap = 3  # Existing donloaded data will be redownloaded and overwritten if it is within this number of days of now.

    def __get_json(self, page_html, key):
        found = re.search(key + r" = JSON.parse\(\"(.*)\"\);", page_html, re.M)
        if found:
            json_text = found.group(1).replace('\\"', '"')
            return json.loads(json_text)

    def login(self):
        """Login to Garmin Connect."""
        profile_dir = ConfigManager.get_or_create_fit_files_dir()
        username = self.gc_config.get_user()
        password = self.gc_config.get_password()
        if not username or not password:
            print("Missing config: need username and password. Edit GarminConnectConfig.json.")
            return

        logger.debug("login: %s %s", username, password)
        get_headers = {
            'Referer'                           : self.garmin_connect_login_url
        }
        params = {
            'service'                           : self.modern_rest_client.url(),
            'webhost'                           : self.garmin_connect_base_url,
            'source'                            : self.garmin_connect_login_url,
            'redirectAfterAccountLoginUrl'      : self.modern_rest_client.url(),
            'redirectAfterAccountCreationUrl'   : self.modern_rest_client.url(),
            'gauthHost'                         : self.sso_rest_client.url(),
            'locale'                            : 'en_US',
            'id'                                : 'gauth-widget',
            'cssUrl'                            : self.garmin_connect_css_url,
            'privacyStatementUrl'               : '//connect.garmin.com/en-US/privacy/',
            'clientId'                          : 'GarminConnect',
            'rememberMeShown'                   : 'true',
            'rememberMeChecked'                 : 'false',
            'createAccountShown'                : 'true',
            'openCreateAccount'                 : 'false',
            'displayNameShown'                  : 'false',
            'consumeServiceTicket'              : 'false',
            'initialFocus'                      : 'true',
            'embedWidget'                       : 'false',
            'generateExtraServiceTicket'        : 'true',
            'generateTwoExtraServiceTickets'    : 'false',
            'generateNoServiceTicket'           : 'false',
            'globalOptInShown'                  : 'true',
            'globalOptInChecked'                : 'false',
            'mobile'                            : 'false',
            'connectLegalTerms'                 : 'true',
            'locationPromptShown'               : 'true',
            'showPassword'                      : 'true'
        }
        try:
            response = self.sso_rest_client.get(self.garmin_connect_sso_login, get_headers, params)
        except RestResponseException as e:
            root_logger.error("Exception during login get: %s", e)
            RestClient.save_binary_file('login_get.html', e.response)
            return False
        found = re.search(r"name=\"_csrf\" value=\"(\w*)", response.text, re.M)
        if not found:
            logger.error("_csrf not found: %s", response.status_code)
            RestClient.save_binary_file('login_get.html', response)
            return False
        logger.debug("_csrf found (%s).", found.group(1))

        data = {
            'username'  : username,
            'password'  : password,
            'embed'     : 'false',
            '_csrf'     : found.group(1)
        }
        post_headers = {
            'Referer'       : response.url,
            'Content-Type'  : 'application/x-www-form-urlencoded'
        }
        try:
            response = self.sso_rest_client.post(self.garmin_connect_sso_login, post_headers, params, data)
        except RestException as e:
            root_logger.error("Exception during login post: %s", e)
            return False
        found = re.search(r"\?ticket=([\w-]*)", response.text, re.M)
        if not found:
            logger.error("Login ticket not found (%d).", response.status_code)
            RestClient.save_binary_file('login_post.html', response)
            return False
        params = {
            'ticket' : found.group(1)
        }
        try:
            response = self.modern_rest_client.get('', params=params)
        except RestException:
            logger.error("Login get homepage failed (%d).", response.status_code)
            RestClient.save_binary_file('login_home.html', response)
            return False
        self.user_prefs = self.__get_json(response.text, 'VIEWER_USERPREFERENCES')
        if profile_dir:
            self.modern_rest_client.save_json_to_file(f'{profile_dir}/profile.json', self.user_prefs)
        self.display_name = self.user_prefs['displayName']
        self.social_profile = self.__get_json(response.text, 'VIEWER_SOCIAL_PROFILE')
        self.full_name = self.social_profile['fullName']
        root_logger.info("login: %s (%s)", self.full_name, self.display_name)
        return True

    def __unzip_files(self, outdir):
        """Unzip and downloaded zipped files into the directory supplied."""
        root_logger.info("unzip_files: from %s to %s", self.temp_dir, outdir)
        for filename in os.listdir(self.temp_dir):
            match = re.search(r'.*\.zip', filename)
            if match:
                full_pathname = f'{self.temp_dir}/{filename}'
                with zipfile.ZipFile(full_pathname, 'r') as files_zip:
                    try:
                        files_zip.extractall(outdir)
                    except Exception as e:
                        logger.error('Failed to unzip %s to %s: %s', full_pathname, outdir, e)

    def __get_stat(self, stat_function, directory, date, days, overwite):
        for day in tqdm(range(0, days), unit='days'):
            download_date = date + datetime.timedelta(days=day)
            # always overwrite for yesterday and today since the last download may have been a partial result
            delta = datetime.datetime.now().date() - download_date
            stat_function(directory, download_date, overwite or delta.days <= self.download_days_overlap)
            # pause for a second between every page access
            time.sleep(1)

    def __get_summary_day(self, directory_func, date, overwite=False):
        root_logger.info("get_summary_day: %s", date)
        date_str = date.strftime('%Y-%m-%d')
        params = {
            'calendarDate': date_str,
            '_': str(conversions.dt_to_epoch_ms(conversions.date_to_dt(date)))
        }
        url = f'{self.garmin_connect_daily_summary_url}/{self.display_name}'
        json_filename = f'{directory_func(date.year)}/daily_summary_{date_str}'
        try:
            self.modern_rest_client.download_json_file(url, json_filename, overwite, params)
        except RestException as e:
            root_logger.error("Exception geting daily summary: %s", e)

    def get_daily_summaries(self, directory_func, date, days, overwite):
        """Download the daily summary data from Garmin Connect and save to a JSON file."""
        root_logger.info("Geting daily summaries: %s (%d)", date, days)
        self.__get_stat(self.__get_summary_day, directory_func, date, days, overwite)

    def __get_monitoring_day(self, date):
        root_logger.info("get_monitoring_day: %s to %s", date, self.temp_dir)
        zip_filename = f'{self.temp_dir}/{date}.zip'
        url = f'wellness/{date.strftime("%Y-%m-%d")}'
        try:
            self.download_service_rest_client.download_binary_file(url, zip_filename)
        except RestException as e:
            root_logger.error("Exception geting daily summary: %s", e)

    def get_monitoring(self, directory_func, date, days):
        """Download the daily monitoring data from Garmin Connect, unzip and save the raw files."""
        root_logger.info("Geting monitoring: %s (%d)", date, days)
        for day in tqdm(range(0, days + 1), unit='days'):
            day_date = date + datetime.timedelta(day)
            self.temp_dir = tempfile.mkdtemp()
            self.__get_monitoring_day(day_date)
            self.__unzip_files(directory_func(day_date.year))
            # pause for a second between every page access
            time.sleep(1)

    def __get_weight_day(self, directory, day, overwite=False):
        root_logger.info("Checking weight: %s overwite %r", day, overwite)
        date_str = day.strftime('%Y-%m-%d')
        params = {
            'startDate' : date_str,
            'endDate'   : date_str,
            '_'         : str(conversions.dt_to_epoch_ms(conversions.date_to_dt(day)))
        }
        json_filename = f'{directory}/weight_{date_str}'
        try:
            self.modern_rest_client.download_json_file(self.garmin_connect_weight_url, json_filename, overwite, params)
        except RestException as e:
            root_logger.error("Exception geting daily summary: %s", e)

    def get_weight(self, directory, date, days, overwite):
        """Download the sleep data from Garmin Connect and save to a JSON file."""
        root_logger.info("Geting weight: %s (%d)", date, days)
        self.__get_stat(self.__get_weight_day, directory, date, days, overwite)

    def __get_activity_summaries(self, start, count):
        root_logger.info("get_activity_summaries")
        params = {
            'start' : str(start),
            "limit" : str(count)
        }
        try:
            response = self.modern_rest_client.get(self.garmin_connect_activity_search_url, params=params)
            return response.json()
        except RestException as e:
            root_logger.error("Exception geting activity summary: %s", e)

    def __save_activity_details(self, directory, activity_id_str, overwite):
        root_logger.debug("save_activity_details")
        json_filename = f'{directory}/activity_details_{activity_id_str}'
        try:
            self.activity_service_rest_client.download_json_file(activity_id_str, json_filename, overwite)
        except RestException as e:
            root_logger.error("Exception geting daily summary %s", e)

    def __save_activity_file(self, activity_id_str):
        root_logger.debug("save_activity_file: %s", activity_id_str)
        zip_filename = f'{self.temp_dir}/activity_{activity_id_str}.zip'
        url = f'activity/{activity_id_str}'
        try:
            self.download_service_rest_client.download_binary_file(url, zip_filename)
        except RestException as e:
            root_logger.error("Exception downloading activity file: %s", e)

    def get_activities(self, directory, count, overwite=False):
        """Download activities files from Garmin Connect and save the raw files."""
        self.temp_dir = tempfile.mkdtemp()
        logger.info("Geting activities: '%s' (%d) temp %s", directory, count, self.temp_dir)
        activities = self.__get_activity_summaries(0, count)
        for activity in tqdm(activities or [], unit='activities'):
            activity_id_str = str(activity['activityId'])
            activity_name_str = conversions.printable(activity['activityName'])
            root_logger.info("get_activities: %s (%s)", activity_name_str, activity_id_str)
            json_filename = f'{directory}/activity_{activity_id_str}.json'
            if not os.path.isfile(json_filename) or overwite:
                root_logger.info("get_activities: %s <- %r", json_filename, activity)
                self.__save_activity_details(directory, activity_id_str, overwite)
                self.modern_rest_client.save_json_to_file(json_filename, activity)
                if not os.path.isfile(f'{directory}/{activity_id_str}.fit') or overwite:
                    self.__save_activity_file(activity_id_str)
                # pause for a second between every page access
                time.sleep(1)
        self.__unzip_files(directory)

    def get_activity_types(self, directory, overwite):
        """Download the activity types from Garmin Connect and save to a JSON file."""
        root_logger.info("get_activity_types: '%s'", directory)
        json_filename = f'{directory}/activity_types'
        try:
            self.activity_service_rest_client.download_json_file('activityTypes', json_filename, overwite)
        except RestException as e:
            root_logger.error("Exception geting activity types: %s", e)

    def __get_sleep_day(self, directory, date, overwite=False):
        json_filename = f'{directory}/sleep_{date}'
        params = {
            'date'                  : date.strftime("%Y-%m-%d"),
            'nonSleepBufferMinutes' : 60
        }
        url = f'{self.garmin_connect_sleep_daily_url}/{self.display_name}'
        try:
            self.modern_rest_client.download_json_file(url, json_filename, overwite, params)
        except RestException as e:
            root_logger.error("Exception geting daily summary: %s", e)

    def get_sleep(self, directory, date, days, overwite):
        """Download the sleep data from Garmin Connect and save to a JSON file."""
        root_logger.info("Geting sleep: %s (%d)", date, days)
        self.__get_stat(self.__get_sleep_day, directory, date, days, overwite)

    def __get_rhr_day(self, directory, day, overwite=False):
        date_str = day.strftime('%Y-%m-%d')
        json_filename = f'{directory}/rhr_{date_str}'
        params = {
            'fromDate'  : date_str,
            'untilDate' : date_str,
            'metricId'  : 60
        }
        url = f'{self.garmin_connect_rhr}/{self.display_name}'
        try:
            self.modern_rest_client.download_json_file(url, json_filename, overwite, params)
        except RestException as e:
            root_logger.error("Exception geting daily summary %s", e)

    def get_rhr(self, directory, date, days, overwite):
        """Download the resting heart rate data from Garmin Connect and save to a JSON file."""
        root_logger.info("Geting rhr: %s (%d)", date, days)
        self.__get_stat(self.__get_rhr_day, directory, date, days, overwite)

    def __get_hydration_day(self, directory_func, day, overwite=False):
        date_str = day.strftime('%Y-%m-%d')
        json_filename = f'{directory_func(day.year)}/hydration_{date_str}'
        url = f'{self.garmin_connect_daily_hydration_url}/{date_str}'
        try:
            self.modern_rest_client.download_json_file(url, json_filename, overwite)
        except RestException as e:
            root_logger.error("Exception geting hydration: %s", e)

    def get_hydration(self, directory_func, date, days, overwite):
        """Download the hydration data from Garmin Connect and save to a JSON file."""
        root_logger.info("Geting hydration: %s (%d)", date, days)
        self.__get_stat(self.__get_hydration_day, directory_func, date, days, overwite)



"""Class that encapsulates config data for the application."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


class Config():
    """Class that encapsilates config data for the application."""

    db = {
        'type'                  : 'sqlite'
    }

    directories = {
        'relative_to_home'      : True,
        'config_dir'            : '.GarminDb',
        'base_dir'              : 'HealthData',
        'plugins_dir'           : "Plugins",
        'fit_file_dir'          : 'FitFiles',
        'fitbit_file_dir'       : 'FitBitFiles',
        'mshealth_file_dir'     : 'MSHealth',
        'db_dir'                : 'DBs',
        'backup_dir'            : 'Backups',
        'sleep_files_dir'       : 'Sleep',
        'activities_file_dir'   : 'Activities',
        'monitoring_file_dir'   : 'Monitoring',
        'weight_files_dir'      : 'Weight',
        'rhr_files_dir'         : 'RHR'
    }

    config = {
        'metric'                : False
    }

    device_directories = {
        'base'                  : 'garmin',
        'activities'            : 'activity',
        'monitoring'            : 'monitor',
        'sleep'                 : 'sleep',
        'settings'              : 'settings'
    }

    graphs = {
        'size'                  : [16.0, 12.0],
        'steps'                 : {'period' : 'weeks', 'days' : 730},
        'hr'                    : {'period' : 'weeks', 'days' : 730},
        'itime'                 : {'period' : 'weeks', 'days' : 730},
        'weight'                : {'period' : 'weeks', 'days' : 730}
    }

    checkup = {
        'look_back_days'        : 90
    }



"""A script for importing CSV formatted FitBit export data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import logging
from tqdm import tqdm

from idbutils import CsvImporter
from idbutils import FileProcessor

from .fitbit_db import FitBitDb, DaysSummary


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))


class FitBitData(object):
    """A object for importing CSV formatted FitBit export data."""

    cols_map = {
        'sleep-minutesAwake': ('awake_mins', CsvImporter.map_integer),
        'activities-caloriesBMR': ('calories_bmr', CsvImporter.map_integer),
        'sleep-minutesToFallAsleep': ('to_fall_asleep_mins', CsvImporter.map_integer),
        'activities-floors': ('floors', CsvImporter.map_integer),
        'activities-steps': ('steps', CsvImporter.map_integer),
        'activities-distance': ('distance', CsvImporter.map_float),
        'foods-log-caloriesIn': ('calories_in', CsvImporter.map_integer),
        'activities-activityCalories': ('activities_calories', CsvImporter.map_integer),
        'sleep-minutesAfterWakeup': ('after_wakeup_mins', CsvImporter.map_integer),
        'activities-minutesFairlyActive': ('fairly_active_mins', CsvImporter.map_integer),
        'sleep-efficiency': ('sleep_efficiency', CsvImporter.map_integer),
        'sleep-timeInBed': ('in_bed_mins', CsvImporter.map_integer),
        'activities-minutesVeryActive': ('very_active_mins', CsvImporter.map_integer),
        'body-weight': ('weight', CsvImporter.map_kgs),
        'activities-minutesSedentary': ('sedentary_mins', CsvImporter.map_integer),
        'activities-elevation': ('elevation', CsvImporter.map_meters),
        'activities-minutesLightlyActive': ('lightly_active_mins', CsvImporter.map_integer),
        'sleep-startTime': ('sleep_start', CsvImporter.map_time),
        'activities-calories': ('calories', CsvImporter.map_integer),
        'foods-log-water': ('log_water', CsvImporter.map_float),
        'sleep-minutesAsleep': ('asleep_mins', CsvImporter.map_integer),
        'body-bmi': ('bmi', CsvImporter.map_float),
        'dateTime': ('day', CsvImporter.map_ymd_date),
        'body-fat': ('body_fat', CsvImporter.map_float),
        'sleep-awakeningsCount': ('awakenings_count', CsvImporter.map_integer),
    }

    def __init__(self, input_file, input_dir, db_params, metric, debug):
        """Return a new instance of FitBitData given the location of the data files, paramters for accessing the database, and if the data should be stored in metric units."""
        self.metric = metric
        self.fitbitdb = FitBitDb(db_params, debug)
        if input_file:
            self.file_names = FileProcessor.match_file(input_file, r'.*\.csv')
        if input_dir:
            self.file_names = FileProcessor.dir_to_files(input_dir, r'.*\.csv')

    def file_count(self):
        """Return the number of files that will be propcessed."""
        return len(self.file_names)

    def __write_entry(self, db_entry):
        DaysSummary.insert_or_update(self.fitbitdb, DaysSummary.intersection(db_entry))

    def process_files(self):
        """Import files into a database."""
        for file_name in tqdm(self.file_names, unit='files'):
            logger.info("Processing file: " + file_name)
            self.csvimporter = CsvImporter(file_name, self.cols_map, self.__write_entry)
            self.csvimporter.process_file(not self.metric)



"""Objects representing a database and database objects for storing health summary data from a Garmin device."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
from sqlalchemy import Column, Integer, Date, DateTime

import idbutils

from ..summarydb import SummaryBase


logger = logging.getLogger(__name__)

GarminSummaryDb = idbutils.DB.create('garmin_summary', 8, "Database for storing health summary data from a Garmin device.")
Summary = idbutils.DbObject.create('summary', GarminSummaryDb, 1, base=idbutils.KeyValueObject)


class YearsSummary(GarminSummaryDb.Base, SummaryBase):
    """A table holding summarized data with one row per year."""

    __tablename__ = 'years_summary'

    db = GarminSummaryDb
    table_version = 4
    view_version = SummaryBase.view_version

    first_day = Column(Date, primary_key=True)

    @classmethod
    def create_view(cls, db):
        """Create the default database view for the table."""
        cls.create_years_view(db)


class MonthsSummary(GarminSummaryDb.Base, SummaryBase):
    """A table holding summarized data with one row per month."""

    __tablename__ = 'months_summary'

    db = GarminSummaryDb
    table_version = 4
    view_version = SummaryBase.view_version

    first_day = Column(Date, primary_key=True)

    @classmethod
    def create_view(cls, db):
        """Create the default database view for the table."""
        cls.create_months_view(db)


class WeeksSummary(GarminSummaryDb.Base, SummaryBase):
    """A table holding summarizzed data with one row per week."""

    __tablename__ = 'weeks_summary'

    db = GarminSummaryDb
    table_version = 4
    view_version = SummaryBase.view_version

    first_day = Column(Date, primary_key=True)

    @classmethod
    def create_view(cls, db):
        """Create the default database view for the table."""
        cls.create_weeks_view(db)


class DaysSummary(GarminSummaryDb.Base, SummaryBase):
    """A table holding summarized data with one row per day."""

    __tablename__ = 'days_summary'

    db = GarminSummaryDb
    table_version = 4
    view_version = SummaryBase.view_version

    day = Column(Date, primary_key=True)

    @classmethod
    def create_view(cls, db):
        """Create the default database view for the table."""
        cls.create_days_view(db)


class IntensityHR(GarminSummaryDb.Base, idbutils.DbObject):
    """Monitoring heart rate values that fall within a intensity period."""

    __tablename__ = 'intensity_hr'

    db = GarminSummaryDb
    table_version = 1

    timestamp = Column(DateTime, primary_key=True)
    intensity = Column(Integer, nullable=False)
    heart_rate = Column(Integer, nullable=False)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dictionary of aggregate statistics for the given time period."""
        return {
            'inactive_hr_avg' : cls.s_get_col_avg_for_value(session, cls.heart_rate, cls.intensity, 0, start_ts, end_ts, True),
            'inactive_hr_min' : cls.s_get_col_min_for_value(session, cls.heart_rate, cls.intensity, 0, start_ts, end_ts, True),
            'inactive_hr_max' : cls.s_get_col_max_for_value(session, cls.heart_rate, cls.intensity, 0, start_ts, end_ts, True),
        }



"""Test Garmin database data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging
# from sqlalchemy.exc import LookupError

import fitfile

from garmindb import ConfigManager
from garmindb.garmindb import GarminDb, File, Attributes


root_logger = logging.getLogger()
handler = logging.FileHandler('garmin_db_objects.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestGarminDbObjects(unittest.TestCase):
    """Class for testing Garmin database data."""

    @classmethod
    def setUpClass(cls):
        cls.db_params = ConfigManager.get_db_params(test_db=True)
        cls.garmin_db = GarminDb(cls.db_params)

    def check_file_obj(self, filename_with_path, file_type, file_serial_number):
        (file_id, file_name) = File.name_and_id_from_path(filename_with_path)
        file_dict = {
            'id'            : file_id,
            'name'          : file_name,
            'type'          : file_type,
            'serial_number' : file_serial_number,
        }
        logger.info("check_file_table: %r", file_dict)
        File.insert_or_update(self.garmin_db, file_dict)
        file = File.get(self.garmin_db, file_id)
        self.assertEqual(file.id, file_dict['id'])
        self.assertEqual(file.name, file_dict['name'])
        self.assertEqual(file.type, file_dict['type'])
        self.assertEqual(file.serial_number, file_dict['serial_number'])

    def test_file_good(self):
        file_id = 123345678
        filename = '%d.fit' % file_id
        filename_with_path = '/test/directory/' + filename
        file_type = File.FileType.fit_goals
        file_serial_number = 987654321
        self.check_file_obj(filename_with_path, file_type, file_serial_number)

    def test_file_good_with_activity_name(self):
        file_id = 123345678
        filename = '%d_ACTIVITY.fit' % file_id
        filename_with_path = '/test/directory/' + filename
        file_type = File.FileType.fit_goals
        file_serial_number = 987654321
        self.check_file_obj(filename_with_path, file_type, file_serial_number)

    def test_file_good_with_watch_name(self):
        file_id = 'SBK82515'
        filename = '%s.fit' % file_id
        filename_with_path = '/test/directory/' + filename
        file_type = File.FileType.fit_goals
        file_serial_number = 987654321
        self.check_file_obj(filename_with_path, file_type, file_serial_number)

    def test_file_xbad_type(self):
        file_id = 123345678
        filename = '%d.fit' % file_id
        filename_with_path = '/test/directory/' + filename
        file_type = 'xxxxx'
        file_serial_number = 987654321
        with self.assertRaises(LookupError):
            self.check_file_obj(filename_with_path, file_type, file_serial_number)

    def test_file_type(self):
        file_types_list = list(File.FileType)
        self.assertIn(File.FileType.convert(fitfile.FileType.goals), file_types_list)

    def test_get_default(self):
        result = Attributes.get(self.garmin_db, 'test_String')
        self.assertEqual(result, None)
        result = Attributes.get(self.garmin_db, 'test_String', 'default_value')
        self.assertEqual(result, 'default_value')
        #
        result = Attributes.get_string(self.garmin_db, 'test_String')
        self.assertEqual(result, None)
        result = Attributes.get_string(self.garmin_db, 'test_String', 'default_value')
        self.assertEqual(result, 'default_value')

    def test_set_get_default(self):
        Attributes.set(self.garmin_db, 'test_String', 'test_value')
        result = Attributes.get_string(self.garmin_db, 'test_String')
        self.assertEqual(result, 'test_value')
        result = Attributes.get_string(self.garmin_db, 'test_String', 'default_value')
        self.assertEqual(result, 'test_value')
        #
        Attributes.set_if_unset(self.garmin_db, 'test_String', 'test_value2')
        result = Attributes.get_string(self.garmin_db, 'test_String')
        self.assertEqual(result, 'test_value')

    def test_get_float(self):
        Attributes.set(self.garmin_db, 'test_String', 2.2)
        result = Attributes.get_float(self.garmin_db, 'test_String')
        self.assertEqual(result, 2.2)
        result = Attributes.get_float(self.garmin_db, 'test_String', 'default_value')
        self.assertEqual(result, 2.2)
        #
        Attributes.set_if_unset(self.garmin_db, 'test_String', 2.2)
        result = Attributes.get_float(self.garmin_db, 2.2)

    def test_measurement_system(self):
        result = Attributes.measurements_type(self.garmin_db, fitfile.field_enums.DisplayMeasure.metric)
        self.assertEqual(result, fitfile.field_enums.DisplayMeasure.metric)
        for value in fitfile.field_enums.DisplayMeasure:
            Attributes.set(self.garmin_db, 'measurement_system', value)
            result = Attributes.measurements_type(self.garmin_db)
            self.assertEqual(result, value)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Objects representing a database and database objects for storing health data from a Garmin device."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os
import datetime
import logging
import re
from sqlalchemy import Column, Integer, Date, DateTime, Time, Float, String, Enum, ForeignKey, func, PrimaryKeyConstraint
from sqlalchemy.ext.hybrid import hybrid_property

import fitfile
import idbutils


logger = logging.getLogger(__name__)


class GarminDbError(Exception):
    """Base exception for GarminDb exceptions"""


class GarminDbError_IdNotFound(GarminDbError):
    """File id not found"""


GarminDb = idbutils.DB.create('garmin', 14, "Database for storing health data from a Garmin device.")


class Attributes(GarminDb.Base, idbutils.KeyValueObject):
    """Object representing generic key-value data from a Garmin device."""

    __tablename__ = 'attributes'

    db = GarminDb
    table_version = 1

    @classmethod
    def measurements_type(cls, db, default=None):
        """Return the database units type (metric, statute, etc)."""
        return fitfile.field_enums.DisplayMeasure.from_string(cls.get_string(db, 'measurement_system', default))

    @classmethod
    def measurements_type_metric(cls, db):
        """Return True if the database units are metric."""
        return (cls.measurements_type(db) == fitfile.field_enums.DisplayMeasure.metric)


class Device(GarminDb.Base, idbutils.DbObject):
    """Class representing a Garmin device."""

    __tablename__ = 'devices'

    db = GarminDb
    table_version = 4
    unknown_device_serial_number = 9999999999

    Manufacturer = idbutils.derived_enum.derive('Manufacturer', fitfile.Manufacturer, {'Microsoft' : 100001, 'Unknown': 100000})

    serial_number = Column(Integer, primary_key=True)
    timestamp = Column(DateTime)
    device_type = Column(String)
    manufacturer = Column(Enum(Manufacturer))
    product = Column(String)
    hardware_version = Column(String)

    @property
    def product_as_enum(self):
        """Convert the product attribute from a string to an enum and return it."""
        return fitfile.product_enum(self.manufacturer, self.product)

    @classmethod
    def local_device_serial_number(cls, serial_number, device_type):
        """Return a synthetic serial number for a sub device composed of the parent's serial number and the sub device type."""
        return '%s%06d' % (serial_number, device_type.value)


class DeviceInfo(GarminDb.Base, idbutils.DbObject):
    """Class representing a Garmin device info message from a FIT file."""

    __tablename__ = 'device_info'

    db = GarminDb
    table_version = 4
    view_version = 6

    timestamp = Column(DateTime, nullable=False)
    file_id = Column(String, ForeignKey('files.id'))
    serial_number = Column(Integer, ForeignKey('devices.serial_number'), nullable=False)
    software_version = Column(String)
    cum_operating_time = Column(Time, nullable=False, default=datetime.time.min)
    battery_status = Column(Enum(fitfile.field_enums.BatteryStatus))
    battery_voltage = Column(Float)

    __table_args__ = (
        PrimaryKeyConstraint('timestamp', 'serial_number'),
    )

    @classmethod
    def s_get_from_dict(cls, session, values_dict):
        """Return a single DeviceInfo instance for the given id."""
        return session.query(cls).filter(cls.timestamp == values_dict['timestamp']).filter(cls.serial_number == values_dict['serial_number']).one_or_none()

    @classmethod
    def create_view(cls, db):
        """Create a database view that presents the device info data in a more user friendly way."""
        cols = [
            cls.timestamp.label('timestamp'),
            cls.file_id.label('file_id'),
            cls.serial_number.label('serial_number'),
            Device.device_type.label('device_type'),
            cls.software_version.label('software_version'),
            Device.manufacturer.label('manufacturer'),
            Device.product.label('product'),
            Device.hardware_version.label('hardware_version'),
            cls.battery_status.label('battery_status')
        ]
        cls.create_join_view(db, cls._get_default_view_name(), cols, Device, order_by=cls.timestamp.desc())


class File(GarminDb.Base, idbutils.DbObject):
    """Class representing a data file."""

    __tablename__ = 'files'

    db = GarminDb
    table_version = 3
    view_version = 4

    fit_file_types_prefix = 'fit_'
    FileType = idbutils.derived_enum.derive('FileType', fitfile.FileType, {'tcx' : 100001, 'gpx' : 100002}, fit_file_types_prefix)

    id = Column(String, primary_key=True)
    name = Column(String, unique=True)
    type = Column(Enum(FileType), nullable=False)
    serial_number = Column(Integer, ForeignKey('devices.serial_number'))

    @classmethod
    def s_get_id(cls, session, pathname):
        """Return the id of a file given it's pathname."""
        return cls.s_find_id(session, {File.name: os.path.basename(pathname)})

    @classmethod
    def create_view(cls, db):
        """Create a databse view that presents the file data in a more user friendly way."""
        cols = [
            DeviceInfo.timestamp.label('timestamp'),
            cls.id.label('activity_id'),
            cls.name.label('name'),
            cls.type.label('type'),
            Device.manufacturer.label('manufacturer'),
            Device.product.label('product'),
            Device.serial_number.label('serial_number')
        ]
        cls.create_multi_join_view(db, cls._get_default_view_name(), cols,
                                   [(Device, File.serial_number == Device.serial_number), (DeviceInfo, File.id == DeviceInfo.file_id)],
                                   DeviceInfo.timestamp.desc())

    @classmethod
    def name_and_id_from_path(cls, pathname):
        """Return the name and id of a file given it's pathname."""
        filename = os.path.basename(pathname)
        # first check for file name formats like 123456789_ACTIVITY.fit and 123456789.fit from Garmin Connect
        found = re.match(r"(\d+).*\.\w+", filename)
        if found:
            return (found.group(1), filename)
        # Check for files from a watch with names like SBK82515.FIT
        found = re.match(r"(.+)\.\w+", filename)
        if found:
            return (found.group(1), filename)
        raise GarminDbError_IdNotFound()

    @classmethod
    def id_from_path(cls, pathname):
        """Return the id of a file given it's pathname."""
        id, _ = cls.name_and_id_from_path(pathname)
        return id


class Weight(GarminDb.Base, idbutils.DbObject):
    """Class representing a weight entry."""

    __tablename__ = 'weight'

    db = GarminDb
    table_version = 1

    day = Column(Date, primary_key=True)
    weight = Column(Float, nullable=False)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dictionary of aggregate statistics for the given time period."""
        return {
            'weight_avg': cls.s_get_col_avg(session, cls.weight, start_ts, end_ts, True),
            'weight_min': cls.s_get_col_min(session, cls.weight, start_ts, end_ts, True),
            'weight_max': cls.s_get_col_max(session, cls.weight, start_ts, end_ts)
        }


class Stress(GarminDb.Base, idbutils.DbObject):
    """Class representing a stress reading."""

    __tablename__ = 'stress'

    db = GarminDb
    table_version = 1

    timestamp = Column(DateTime, primary_key=True, unique=True)
    stress = Column(Integer, nullable=False)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dictionary of aggregate statistics for the given time period."""
        return {
            'stress_avg': cls.s_get_col_avg(session, cls.stress, start_ts, end_ts, True),
        }


class Sleep(GarminDb.Base, idbutils.DbObject):
    """Class representing a sleep session."""

    __tablename__ = 'sleep'

    db = GarminDb
    table_version = 1

    day = Column(Date, primary_key=True)
    start = Column(DateTime)
    end = Column(DateTime)
    total_sleep = Column(Time, nullable=False, default=datetime.time.min)
    deep_sleep = Column(Time, nullable=False, default=datetime.time.min)
    light_sleep = Column(Time, nullable=False, default=datetime.time.min)
    rem_sleep = Column(Time, nullable=False, default=datetime.time.min)
    awake = Column(Time, nullable=False, default=datetime.time.min)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dictionary of aggregate statistics for the given time period."""
        return {
            'sleep_avg'     : cls.s_get_time_col_avg(session, cls.total_sleep, start_ts, end_ts),
            'sleep_min'     : cls.s_get_time_col_min(session, cls.total_sleep, start_ts, end_ts),
            'sleep_max'     : cls.s_get_time_col_max(session, cls.total_sleep, start_ts, end_ts),
            'rem_sleep_avg' : cls.s_get_time_col_avg(session, cls.rem_sleep, start_ts, end_ts),
            'rem_sleep_min' : cls.s_get_time_col_min(session, cls.rem_sleep, start_ts, end_ts),
            'rem_sleep_max' : cls.s_get_time_col_max(session, cls.rem_sleep, start_ts, end_ts),
        }


class SleepEvents(GarminDb.Base, idbutils.DbObject):
    """Table that stores events recorded druing sleep."""

    __tablename__ = 'sleep_events'

    db = GarminDb
    table_version = 2

    timestamp = Column(DateTime, primary_key=True)
    event = Column(String)
    duration = Column(Time, nullable=False, default=datetime.time.min)

    @classmethod
    def get_wake_time(cls, db, day_date):
        """Return the wake time for a given date."""
        day_start_ts = datetime.datetime.combine(day_date, datetime.time.min)
        day_stop_ts = datetime.datetime.combine(day_date, datetime.time.max)
        values = cls.get_col_values(db, cls.timestamp, cls.event, 'wake_time', day_start_ts, day_stop_ts)
        if len(values) > 0:
            return values[0][0]


class RestingHeartRate(GarminDb.Base, idbutils.DbObject):
    """Class representing a daily resting heart rate reading."""

    __tablename__ = 'resting_hr'

    db = GarminDb
    table_version = 1
    _col_units = {'resting_heart_rate': 'bpm'}

    day = Column(Date, primary_key=True)
    resting_heart_rate = Column(Float)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dictionary of aggregate statistics for the given time period."""
        return {
            'rhr_avg': cls.s_get_col_avg(session, cls.resting_heart_rate, start_ts, end_ts, ignore_le_zero=True),
            'rhr_min': cls.s_get_col_min(session, cls.resting_heart_rate, start_ts, end_ts, ignore_le_zero=True),
            'rhr_max': cls.s_get_col_max(session, cls.resting_heart_rate, start_ts, end_ts),
        }


class DailySummary(GarminDb.Base, idbutils.DbObject):
    """Class representing a Garmin daily summary."""

    __tablename__ = 'daily_summary'

    db = GarminDb
    table_version = 4
    _col_units = {'hr_min': 'bpm', 'hr_max': 'bpm', 'rhr': 'bpm'}

    day = Column(Date, primary_key=True)
    hr_min = Column(Integer)
    hr_max = Column(Integer)
    rhr = Column(Integer)
    stress_avg = Column(Integer)
    step_goal = Column(Integer)
    steps = Column(Integer)
    moderate_activity_time = Column(Time, nullable=False, default=datetime.time.min)
    vigorous_activity_time = Column(Time, nullable=False, default=datetime.time.min)
    intensity_time_goal = Column(Time, nullable=False, default=datetime.time.min)
    floors_up = Column(Float)
    floors_down = Column(Float)
    floors_goal = Column(Float)
    distance = Column(Float)
    calories_goal = Column(Integer)
    calories_total = Column(Integer)
    calories_bmr = Column(Integer)
    calories_active = Column(Integer)
    calories_consumed = Column(Integer)
    hydration_goal = Column(Integer)
    hydration_intake = Column(Integer)
    sweat_loss = Column(Integer)
    spo2_avg = Column(Float)
    spo2_min = Column(Float)
    rr_waking_avg = Column(Float)
    rr_max = Column(Float)
    rr_min = Column(Float)
    description = Column(String)

    @hybrid_property
    def intensity_time(self):
        """Return intensity_time computed from moderate_activity_time and vigorous_activity_time."""
        return fitfile.conversions.add_time(self.moderate_activity_time, self.vigorous_activity_time, 2)

    @intensity_time.expression
    def intensity_time(cls):
        """Return intensity_time computed from moderate_activity_time and vigorous_activity_time."""
        return cls._time_from_secs(2 * cls._secs_from_time(cls.vigorous_activity_time) + cls._secs_from_time(cls.moderate_activity_time))

    @hybrid_property
    def intensity_time_goal_percent(self):
        """Return the percentage of intensity time goal achieved."""
        if self.intensity_time is not None and self.intensity_time_goal is not None:
            return (fitfile.conversions.time_to_secs(self.intensity_time) * 100) / fitfile.conversions.time_to_secs(self.intensity_time_goal)
        return 0.0

    @intensity_time_goal_percent.expression
    def intensity_time_goal_percent(cls):
        """Return the percentage of intensity time goal achieved."""
        return func.round((cls._secs_from_time(cls.intensity_time) * 100) / cls._secs_from_time(cls.intensity_time_goal))

    @hybrid_property
    def steps_goal_percent(self):
        """Return the percentage of steps goal achieved."""
        if self.steps is not None and self.step_goal is not None:
            return (self.steps * 100) / self.step_goal
        return 0.0

    @steps_goal_percent.expression
    def steps_goal_percent(cls):
        """Return the percentage of steps goal achieved."""
        return func.round((cls.steps * 100) / cls.step_goal)

    @hybrid_property
    def floors_goal_percent(self):
        """Return the percentage of floors goal achieved."""
        if self.floors_up is not None and self.floors_goal is not None:
            return (self.floors_up * 100) / self.floors_goal
        return 0.0

    @floors_goal_percent.expression
    def floors_goal_percent(cls):
        """Return the percentage of floors goal achieved."""
        return func.round((cls.floors_up * 100) / cls.floors_goal)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dictionary of aggregate statistics for the given time period."""
        return {
            'rhr_avg'                   : cls.s_get_col_avg(session, cls.rhr, start_ts, end_ts),
            'rhr_min'                   : cls.s_get_col_min(session, cls.rhr, start_ts, end_ts),
            'rhr_max'                   : cls.s_get_col_max(session, cls.rhr, start_ts, end_ts),
            'stress_avg'                : cls.s_get_col_avg(session, cls.stress_avg, start_ts, end_ts),
            'steps'                     : cls.s_get_col_sum(session, cls.steps, start_ts, end_ts),
            'steps_goal'                : cls.s_get_col_sum(session, cls.step_goal, start_ts, end_ts),
            'floors'                    : cls.s_get_col_sum(session, cls.floors_up, start_ts, end_ts),
            'floors_goal'               : cls.s_get_col_sum(session, cls.floors_goal, start_ts, end_ts),
            'intensity_time'            : cls.s_get_time_col_avg(session, cls.intensity_time, start_ts, end_ts),
            'moderate_activity_time'    : cls.s_get_time_col_avg(session, cls.moderate_activity_time, start_ts, end_ts),
            'vigorous_activity_time'    : cls.s_get_time_col_sum(session, cls.vigorous_activity_time, start_ts, end_ts),
            'intensity_time_goal'       : cls.s_get_time_col_avg(session, cls.intensity_time_goal, start_ts, end_ts),
            'calories_goal'             : cls.s_get_col_sum(session, cls.calories_goal, start_ts, end_ts),
            'calories_avg'              : cls.s_get_col_avg(session, cls.calories_total, start_ts, end_ts),
            'calories_bmr_avg'          : cls.s_get_col_avg(session, cls.calories_bmr, start_ts, end_ts),
            'calories_active_avg'       : cls.s_get_col_avg(session, cls.calories_active, start_ts, end_ts),
            'calories_consumed_avg'     : cls.s_get_col_avg(session, cls.calories_consumed, start_ts, end_ts),
            'hydration_goal'            : cls.s_get_col_sum(session, cls.hydration_goal, start_ts, end_ts),
            'hydration_avg'             : cls.s_get_col_avg(session, cls.hydration_intake, start_ts, end_ts),
            'hydration_intake'          : cls.s_get_col_sum(session, cls.hydration_intake, start_ts, end_ts),
            'sweat_loss_avg'            : cls.s_get_col_avg(session, cls.sweat_loss, start_ts, end_ts),
            'sweat_loss'                : cls.s_get_col_sum(session, cls.sweat_loss, start_ts, end_ts),
            'spo2_avg'                  : cls.s_get_col_avg(session, cls.spo2_avg, start_ts, end_ts),
            'spo2_min'                  : cls.s_get_col_min(session, cls.spo2_min, start_ts, end_ts),
            'rr_waking_avg'             : cls.s_get_col_avg(session, cls.rr_waking_avg, start_ts, end_ts),
            'rr_max'                    : cls.s_get_col_max(session, cls.rr_max, start_ts, end_ts),
            'rr_min'                    : cls.s_get_col_min(session, cls.rr_min, start_ts, end_ts),
        }

    @classmethod
    def get_daily_stats(cls, session, day_ts):
        """Return a dictionary of aggregate statistics for the given day."""
        stats = cls.get_stats(session, day_ts, day_ts + datetime.timedelta(1))
        # intensity_time_goal is a weekly goal, so the daily value is 1/7 of the weekly goal
        stats['intensity_time_goal'] = cls._time_from_secs(cls._secs_from_time(stats['intensity_time_goal']) / 7)
        stats['day'] = day_ts
        return stats

    @classmethod
    def get_monthly_stats(cls, session, first_day_ts, last_day_ts):
        """Return a dictionary of aggregate statistics for the given month."""
        stats = cls.get_stats(session, first_day_ts, last_day_ts)
        # intensity time is a weekly goal, so sum up the weekly average values
        first_week_end = first_day_ts + datetime.timedelta(7)
        second_week_end = first_day_ts + datetime.timedelta(14)
        third_week_end = first_day_ts + datetime.timedelta(21)
        fourth_week_end = first_day_ts + datetime.timedelta(28)
        stats['intensity_time_goal'] = fitfile.conversions.add_time(
            fitfile.conversions.add_time(
                cls.s_get_time_col_avg(session, cls.intensity_time_goal, first_day_ts, first_week_end),
                cls.s_get_time_col_avg(session, cls.intensity_time_goal, first_week_end, second_week_end)
            ),
            fitfile.conversions.add_time(
                cls.s_get_time_col_avg(session, cls.intensity_time_goal, second_week_end, third_week_end),
                cls.s_get_time_col_avg(session, cls.intensity_time_goal, third_week_end, fourth_week_end)
            )
        )
        stats['first_day'] = first_day_ts
        return stats



"""Library for storing health metrics in a database."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .summary_base import SummaryBase
from .summary_db import SummaryDb, Summary, YearsSummary, MonthsSummary, WeeksSummary, DaysSummary



"""Objects representing a database and database objects for storing health monitoring data from a Garmin device."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import logging
import datetime
from sqlalchemy import Column, Integer, DateTime, Time, Float, Enum, FLOAT, UniqueConstraint, PrimaryKeyConstraint
from sqlalchemy.ext.hybrid import hybrid_property

import fitfile
import idbutils


logger = logging.getLogger(__name__)

MonitoringDb = idbutils.DB.create('garmin_monitoring', 6, "Database for storing daily health monitoring data from a Garmin device.")


class MonitoringInfo(MonitoringDb.Base, idbutils.DbObject):
    """Class representing data from a health monitoring file."""

    __tablename__ = 'monitoring_info'

    db = MonitoringDb
    table_version = 1

    timestamp = Column(DateTime)
    file_id = Column(Integer, nullable=False)
    activity_type = Column(Enum(fitfile.field_enums.ActivityType))
    resting_metabolic_rate = Column(Integer)
    cycles_to_distance = Column(FLOAT)
    cycles_to_calories = Column(FLOAT)

    __table_args__ = (
        PrimaryKeyConstraint("timestamp", "activity_type"),
    )

    @classmethod
    def s_get_from_dict(cls, session, values_dict):
        """Return a single DeviceInfo instance for the given id."""
        return session.query(cls).filter(cls.timestamp == values_dict['timestamp']).filter(cls.activity_type == values_dict['activity_type']).one_or_none()

    @classmethod
    def get_daily_bmr(cls, db, day_ts):
        """Return the base metabolic rate for the given day."""
        return cls.get_col_avg_of_max_per_day(db, cls.resting_metabolic_rate, day_ts, day_ts + datetime.timedelta(1))

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dict of stats for table entries within the time span."""
        stats = {
            'calories_bmr_avg' : cls.s_get_col_avg(session, cls.resting_metabolic_rate, start_ts, end_ts),
        }
        return stats


class MonitoringHeartRate(MonitoringDb.Base, idbutils.DbObject):
    """Class that reprsents a database table holding resting heart rate data."""

    __tablename__ = 'monitoring_hr'

    db = MonitoringDb
    table_version = 1

    timestamp = Column(DateTime, primary_key=True)
    heart_rate = Column(Integer, nullable=False)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dict of stats for table entries within the time span."""
        return {
            'hr_avg' : cls.s_get_col_avg(session, cls.heart_rate, start_ts, end_ts, True),
            'hr_min' : cls.s_get_col_min(session, cls.heart_rate, start_ts, end_ts, True),
            'hr_max' : cls.s_get_col_max(session, cls.heart_rate, start_ts, end_ts),
        }

    @classmethod
    def get_resting_heartrate(cls, db, wake_ts):
        """Return a resting heart rate value for the day specified."""
        start_ts = wake_ts - datetime.timedelta(0, 0, 0, 0, 10)
        return cls.get_col_min(db, cls.heart_rate, start_ts, wake_ts, True)


class MonitoringIntensity(MonitoringDb.Base, idbutils.DbObject):
    """Class representing monitoring data about cardio minutes."""

    __tablename__ = 'monitoring_intensity'

    db = MonitoringDb
    table_version = 1

    timestamp = Column(DateTime, primary_key=True)
    moderate_activity_time = Column(Time, nullable=False, default=datetime.time.min)
    vigorous_activity_time = Column(Time, nullable=False, default=datetime.time.min)

    __table_args__ = (
        UniqueConstraint("timestamp", "moderate_activity_time", "vigorous_activity_time"),
    )

    @hybrid_property
    def intensity_time(self):
        """Return the total cardio minutes, moderate and vigorous, with vigorous counted double."""
        return fitfile.conversions.add_time(self.moderate_activity_time, self.vigorous_activity_time, 2)

    @intensity_time.expression
    def intensity_time(cls):
        return cls._time_from_secs(2 * cls._secs_from_time(cls.vigorous_activity_time) + cls._secs_from_time(cls.moderate_activity_time))

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dict of stats for table entries within the time span."""
        return {
            'intensity_time'            : cls.s_get_time_col_sum(session, cls.intensity_time, start_ts, end_ts),
            'moderate_activity_time'    : cls.s_get_time_col_sum(session, cls.moderate_activity_time, start_ts, end_ts),
            'vigorous_activity_time'    : cls.s_get_time_col_sum(session, cls.vigorous_activity_time, start_ts, end_ts),
        }


class MonitoringClimb(MonitoringDb.Base, idbutils.DbObject):
    """Class representing monitoring data about elvation gained."""

    __tablename__ = 'monitoring_climb'

    db = MonitoringDb
    table_version = 1

    feet_to_floors = 10
    meters_to_floors = 3

    timestamp = Column(DateTime, primary_key=True)
    # meters or feet
    ascent = Column(Float)
    descent = Column(Float)
    cum_ascent = Column(Float)
    cum_descent = Column(Float)

    __table_args__ = (
        UniqueConstraint("timestamp", "ascent", "descent", "cum_ascent", "cum_descent"),
    )

    @classmethod
    def get_stats(cls, session, func, start_ts, end_ts, measurement_system):
        """Return a dict of stats for table entries within the time span."""
        cum_ascent = func(session, cls.cum_ascent, start_ts, end_ts)
        if cum_ascent:
            if measurement_system is fitfile.field_enums.DisplayMeasure.metric:
                floors = cum_ascent / cls.feet_to_floors
            else:
                floors = cum_ascent / cls.meters_to_floors
        else:
            floors = 0
        return {'floors' : floors}

    @classmethod
    def get_daily_stats(cls, session, day_ts, measurement_system):
        """Return a dict of stats for table entries for the given day."""
        stats = cls.get_stats(session, cls.s_get_col_max, day_ts, day_ts + datetime.timedelta(1), measurement_system)
        stats['day'] = day_ts
        return stats

    @classmethod
    def get_weekly_stats(cls, session, first_day_ts, measurement_system):
        """Return a dict of stats for table entries for the week day."""
        stats = cls.get_stats(session, cls.s_get_col_sum_of_max_per_day, first_day_ts, first_day_ts + datetime.timedelta(7), measurement_system)
        stats['first_day'] = first_day_ts
        return stats

    @classmethod
    def get_monthly_stats(cls, session, first_day_ts, last_day_ts, measurement_system):
        """Return a dict of stats for table entries for the month."""
        stats = cls.get_stats(session, cls.s_get_col_sum_of_max_per_day, first_day_ts, last_day_ts, measurement_system)
        stats['first_day'] = first_day_ts
        return stats

    @classmethod
    def get_yearly_stats(cls, session, year, measurement_system):
        """Return a dict of stats for table entries for the year."""
        first_day_ts = datetime.datetime(year, 1, 1)
        stats = cls.get_stats(session, cls.s_get_col_sum_of_max_per_day, first_day_ts, first_day_ts + datetime.timedelta(365), measurement_system)
        stats['first_day'] = first_day_ts
        return stats


class Monitoring(MonitoringDb.Base, idbutils.DbObject):
    """A table containing monitoring data."""

    __tablename__ = 'monitoring'

    db = MonitoringDb
    table_version = 2

    timestamp = Column(DateTime, nullable=False)
    activity_type = Column(Enum(fitfile.field_enums.ActivityType))
    intensity = Column(Integer)
    duration = Column(Time, nullable=False, default=datetime.time.min)
    distance = Column(Float)
    cum_active_time = Column(Time, nullable=False, default=datetime.time.min)
    active_calories = Column(Integer)
    steps = Column(Integer)
    strokes = Column(Integer)
    cycles = Column(Float)

    __table_args__ = (PrimaryKeyConstraint("timestamp", "activity_type"),)

    @classmethod
    def s_get_from_dict(cls, session, values_dict):
        """Return a single DeviceInfo instance for the given id."""
        return session.query(cls).filter(cls.timestamp == values_dict['timestamp']).filter(cls.activity_type == values_dict['activity_type']).one_or_none()

    @classmethod
    def get_active_calories(cls, session, activity_type, start_ts, end_ts):
        """Return the total calories burned during activity during the indicated period."""
        active_calories = cls.s_get_col_avg_of_max_per_day_for_value(session, cls.active_calories, cls.activity_type, activity_type, start_ts, end_ts)
        return active_calories if active_calories is not None else 0

    @classmethod
    def get_stats(cls, session, func, start_ts, end_ts):
        """Return a dict of stats for table entries within the time span."""
        return {
            'steps': func(session, cls.steps, start_ts, end_ts),
            'calories_active_avg': (
                cls.get_active_calories(session, fitfile.field_enums.ActivityType.running, start_ts, end_ts)
                + cls.get_active_calories(session, fitfile.field_enums.ActivityType.cycling, start_ts, end_ts)
                + cls.get_active_calories(session, fitfile.field_enums.ActivityType.walking, start_ts, end_ts)
            )
        }

    @classmethod
    def get_daily_stats(cls, session, day_ts):
        """Return a dict of stats for table entries for the given day."""
        stats = cls.get_stats(session, cls.s_get_col_max, day_ts, day_ts + datetime.timedelta(1))
        stats['day'] = day_ts
        return stats

    @classmethod
    def get_weekly_stats(cls, session, first_day_ts):
        """Return a dict of stats for table entries for the given week."""
        stats = cls.get_stats(session, cls.s_get_col_sum_of_max_per_day, first_day_ts, first_day_ts + datetime.timedelta(7))
        stats['first_day'] = first_day_ts
        return stats

    @classmethod
    def get_monthly_stats(cls, session, first_day_ts, last_day_ts):
        """Return a dict of stats for table entries for the given week."""
        stats = cls.get_stats(session, cls.s_get_col_sum_of_max_per_day, first_day_ts, last_day_ts)
        stats['first_day'] = first_day_ts
        return stats


class MonitoringRespirationRate(MonitoringDb.Base, idbutils.DbObject):
    """Class that represents a database table holding respiration rate measured in breaths per minute."""

    __tablename__ = 'monitoring_rr'

    db = MonitoringDb
    table_version = 1

    timestamp = Column(DateTime, primary_key=True)
    rr = Column(Float, nullable=False)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dict of stats for table entries within the time span."""
        return {
            'rr_avg' : cls.s_get_col_avg(session, cls.rr, start_ts, end_ts, True),
            'rr_min' : cls.s_get_col_min(session, cls.rr, start_ts, end_ts, True),
            'rr_max' : cls.s_get_col_max(session, cls.rr, start_ts, end_ts),
        }


class MonitoringPulseOx(MonitoringDb.Base, idbutils.DbObject):
    """Class that represents a database table holding pulse ox measurements in percent."""

    __tablename__ = 'monitoring_pulse_ox'

    db = MonitoringDb
    table_version = 1

    timestamp = Column(DateTime, primary_key=True)
    pulse_ox = Column(Float, nullable=False)

    @classmethod
    def get_stats(cls, session, start_ts, end_ts):
        """Return a dict of stats for table entries within the time span."""
        return {
            'pulse_ox_avg' : cls.s_get_col_avg(session, cls.pulse_ox, start_ts, end_ts, True),
            'pulse_ox_min' : cls.s_get_col_min(session, cls.pulse_ox, start_ts, end_ts, True),
            'pulse_ox_max' : cls.s_get_col_max(session, cls.pulse_ox, start_ts, end_ts),
        }



"""Test summary db."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging

from garmindb import summarydb, ConfigManager

from test_summary_db_base import TestSummaryDBBase


root_logger = logging.getLogger()
handler = logging.FileHandler('summary_db.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestSummaryDB(TestSummaryDBBase, unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        db_params = ConfigManager.get_db_params()
        db = summarydb.SummaryDb(db_params)
        table_dict = {
            'summary_table' : summarydb.Summary,
            'months_table' : summarydb.MonthsSummary,
            'weeks_table' : summarydb.WeeksSummary,
            'days_table' : summarydb.DaysSummary
        }
        super().setUpClass(db, table_dict)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Enumeration of types of statistcs that can be downloaded and processed."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import enum


class Statistics(enum.Enum):
    """The types of statistics that can be downloaded and analyzed."""

    monitoring = 1
    steps = 2
    itime = 3
    sleep = 4
    rhr = 5
    weight = 6
    activities = 7

    @classmethod
    def from_string(cls, string):
        """Return a Statistics created from a string that matches an enum name of value."""
        try:
            return cls(string)
        except Exception:
            return getattr(cls, string)



"""Class for importing monitoring FIT files into a database."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import sys
import logging
import traceback
from tqdm import tqdm

import fitfile
from idbutils import FileProcessor


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class FitData(object):
    """Class for importing FIT files into a database."""

    def __init__(self, input_dir, debug, latest=False, recursive=False, fit_types=None, measurement_system=fitfile.field_enums.DisplayMeasure.metric):
        """
        Return an instance of FitData.

        Parameters:
        input_dir (string): directory (full path) to check for monitoring data files
        debug (Boolean): enable debug logging
        latest (Boolean): check for latest files only
        fit_types (Fit.field_enums.FileType): check for this file type only
        measurement_system (enum): which measurement system to use when importing the files

        """
        logger.info("Processing %s FIT data from %s", fit_types, input_dir)
        self.measurement_system = measurement_system
        self.debug = debug
        self.fit_types = fit_types
        self.file_names = FileProcessor.dir_to_files(input_dir, fitfile.file.name_regex, latest, recursive)

    def file_count(self):
        """Return the number of files that will be processed."""
        return len(self.file_names)

    def process_files(self, fit_file_processor):
        """Import FIT files into the database."""
        for file_name in tqdm(self.file_names, unit='files'):
            try:
                fit_file = fitfile.file.File(file_name, self.measurement_system)
                if self.fit_types is None or fit_file.type in self.fit_types:
                    fit_file_processor.write_file(fit_file)
                    root_logger.debug("Wrote %s to the database", fit_file)
                else:
                    root_logger.info("skipping non-matching %s", fit_file)
            except Exception as e:
                logger.error("Failed to parse %s: %s", file_name, e)
                root_logger.error("Failed to parse %s: %s - %s", file_name, e, traceback.format_exc())



"""Test copying FIT files from a USB mounted watch."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging
import datetime

from garmindb import Copy, ConfigManager, GarminConnectConfigManager


root_logger = logging.getLogger()
handler = logging.FileHandler('copy.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestCopy(unittest.TestCase):
    """Class for testing FIT file copying."""

    @classmethod
    def setUpClass(cls):
        cls.gc_config = GarminConnectConfigManager()
        cls.copy = Copy(cls.gc_config.device_mount_dir())

    def test_copy_activity(self):
        activities_dir = ConfigManager.get_or_create_activities_dir(test_dir=True)
        logger.info("Copying activities to %s", activities_dir)
        self.copy.copy_activities(activities_dir)

    def test_copy_monitoring(self):
        monitoring_dir = ConfigManager.get_or_create_monitoring_dir(datetime.datetime.now().year, test_dir=True)
        logger.info("Copying monitoring to %s", monitoring_dir)
        self.copy.copy_monitoring(monitoring_dir)

    def test_copy_settings(self):
        settings_dir = ConfigManager.get_or_create_fit_files_dir(test_dir=True)
        root_logger.info("Copying settings to %s", settings_dir)
        self.copy.copy_settings(settings_dir)

    def test_copy_sleep(self):
        monitoring_dir = ConfigManager.get_or_create_monitoring_dir(datetime.datetime.now().year, test_dir=True)
        root_logger.info("Copying sleep to %s", monitoring_dir)
        self.copy.copy_sleep(monitoring_dir)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""A database and database objects for storing health data from Garmin Connect."""

import re
import os
from setuptools import setup


def get_version(version_file):
    """Extract version fron the module source."""
    with open(version_file, 'r') as file:
        data = file.read()
        match = re.search(r'version_info = \((\d), (\d), (\d)\)', data, re.M)
        if match:
            return f'{match.group(1)}.{match.group(2)}.{match.group(3)}'


def get_long_description(readme_file):
    """Extract long description fron the module readme."""
    with open(readme_file, "r", encoding="utf-8") as file:
        return file.read()


module_name = 'garmindb'
module_version = get_version(module_name + os.sep + 'version_info.py')
module_long_description = get_long_description('README.md')

print(f"Building {module_name} {module_version}\n{module_long_description}")

setup(name=module_name, version=module_version, author='Tom Goetz',
      packages=[module_name, f'{module_name}.garmindb', f'{module_name}.fitbitdb', f'{module_name}.mshealthdb', f'{module_name}.summarydb'],
      scripts=['scripts/garmindb_cli.py', 'scripts/garmindb_graphs.py', 'scripts/garmindb_checkup.py', 'scripts/fitbit.py', 'scripts/mshealth.py'],
      description='Download data from Garmin Connect and store it in a SQLite db for analysis.',
      long_description=module_long_description,
      long_description_content_type='text/markdown',
      url="https://github.com/tcgoetz/GarminDB",
      project_urls={"Bug Tracker": "https://github.com/tcgoetz/GarminDB/issues"},
      install_requires=open('requirements.txt').readlines(),
      classifiers=[
          'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
          "Programming Language :: Python :: 3",
          "Operating System :: OS Independent"
      ],
      python_requires=">=3.0")



"""Class that takes a parsed FIT file object and imports it into a database."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
import sys
import traceback

import fitfile

from .garmindb import GarminDb, File, Device, DeviceInfo, Stress, Attributes


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class FitFileProcessor(object):
    """Class that takes a parsed FIT file object and imports it into a database."""

    def __init__(self, db_params, plugin_manager, debug=0):
        """
        Return a new FitFileProcessor instance.

        Paramters:
        db_params (dict): database access configuration
        debug (Boolean): if True, debug logging is enabled
        """
        root_logger.info("Debug: %s", debug)
        self.plugin_manager = plugin_manager
        self.db_params = db_params
        self.debug = debug
        self.garmin_db = GarminDb(db_params, debug - 1)

    def __write_generic(self, fit_file, message_type, messages):
        """Write all messages of a given message type to the database."""
        handler_name = '_write_' + message_type.name + '_entry'
        function = getattr(self, handler_name, None)
        if function is not None:
            for message in messages:
                try:
                    function(fit_file, message.fields)
                except Exception as e:
                    logger.error("Failed to write message %r type %r: %s", message_type, message, e)
                    root_logger.error("Failed to write message %r type %r: %s", message_type, message, traceback.format_exc())
        elif isinstance(message_type, fitfile.UnknownMessageType) or message_type.is_unknown():
            root_logger.debug("No entry handler %s for message type %r (%d) from %s: %s",
                              handler_name, message_type, len(messages), fit_file.filename, messages[0])
        else:
            root_logger.info("No entry handler %s for known message type %r (%d) from %s: %s",
                             handler_name, message_type, len(messages), fit_file.filename, messages[0])

    def _write_file_id(self, fit_file, message_type, messages):
        """Write all file id messages to the database."""
        self.serial_number = None
        self.manufacturer = None
        self.product = None
        for message in messages:
            self._write_file_id_entry(fit_file, message.fields)

    def __write_message_type(self, fit_file, message_type):
        messages = fit_file[message_type]
        function = getattr(self, '_write_' + message_type.name, self.__write_generic)
        function(fit_file, message_type, messages)
        root_logger.debug("Processed %d %r entries for %s", len(messages), message_type, fit_file.filename)

    def _write_message_types(self, fit_file, message_types):
        """Write all messages from the FIT file to the database ordered by message type."""
        root_logger.info("Importing %s (%s) [%s] with message types: %s", fit_file.filename, fit_file.time_created_local, fit_file.type, message_types)
        #
        # Some ordering is important: 1. create new file entries 2. create new device entries
        #
        priority_message_types = [fitfile.MessageType.file_id, fitfile.MessageType.device_info]
        for message_type in priority_message_types:
            self.__write_message_type(fit_file, message_type)
        for message_type in message_types:
            if message_type not in priority_message_types:
                self.__write_message_type(fit_file, message_type)

    def write_file(self, fit_file):
        """Write all data from the FIT file to database files."""
        with self.garmin_db.managed_session() as self.garmin_db_session:
            self._write_message_types(fit_file, fit_file.message_types)

    #
    # Message type handlers
    #
    def _write_file_id_entry(self, fit_file, message_fields):
        root_logger.debug("file_id fields: %r", message_fields)
        self.serial_number = message_fields.serial_number
        _manufacturer = Device.Manufacturer.convert(message_fields.manufacturer)
        if _manufacturer is not None:
            self.manufacturer = _manufacturer
        self.product = message_fields.product
        device_type = fitfile.MainDeviceType.derive_device_type(self.manufacturer, self.product)
        if self.serial_number:
            device = {
                'serial_number' : self.serial_number,
                'timestamp'     : fit_file.utc_datetime_to_local(message_fields.time_created),
                'device_type'   : fitfile.field_enums.name_for_enum(device_type),
                'manufacturer'  : self.manufacturer,
                'product'       : fitfile.field_enums.name_for_enum(self.product),
            }
            Device.s_insert_or_update(self.garmin_db_session, device)
        (file_id, file_name) = File.name_and_id_from_path(fit_file.filename)
        file = {
            'id'            : file_id,
            'name'          : file_name,
            'type'          : File.FileType.convert(message_fields.type),
            'serial_number' : self.serial_number
        }
        File.s_insert_or_update(self.garmin_db_session, file)

    def _write_device_info_entry(self, fit_file, message_fields):
        timestamp = fit_file.utc_datetime_to_local(message_fields.timestamp)
        device_type = message_fields.get('device_type', fitfile.MainDeviceType.fitness_tracker)
        serial_number = message_fields.serial_number
        manufacturer = Device.Manufacturer.convert(message_fields.manufacturer)
        product = message_fields.product
        source_type = message_fields.source_type
        # local devices are part of the main device. Base missing fields off of the main device.
        if source_type is fitfile.field_enums.SourceType.local:
            if serial_number is None and self.serial_number is not None and device_type is not None:
                serial_number = Device.local_device_serial_number(self.serial_number, device_type)
            if manufacturer is None:
                manufacturer = self.manufacturer
            if product is None:
                product = self.product
        if serial_number is not None:
            device = {
                'serial_number'     : serial_number,
                'timestamp'         : timestamp,
                'device_type'       : fitfile.field_enums.name_for_enum(device_type),
                'manufacturer'      : manufacturer,
                'product'           : fitfile.field_enums.name_for_enum(product),
                'hardware_version'  : message_fields.hardware_version
            }
            Device.s_insert_or_update(self.garmin_db_session, device, ignore_none=True)
            device_info = {
                'file_id'               : File.s_get_id(self.garmin_db_session, fit_file.filename),
                'serial_number'         : serial_number,
                'timestamp'             : timestamp,
                'cum_operating_time'    : message_fields.cum_operating_time,
                'battery_status'        : message_fields.battery_status,
                'battery_voltage'       : message_fields.battery_voltage,
                'software_version'      : message_fields.software_version
            }
            DeviceInfo.s_insert_or_update(self.garmin_db_session, device_info, ignore_none=True)

    def _write_stress_level_entry(self, fit_file, message_fields):
        stress = {
            'timestamp' : message_fields.local_timestamp,
            'stress'    : message_fields.stress_level
        }
        Stress.s_insert_or_update(self.garmin_db_session, stress)

    def _write_event_entry(self, fit_file, message_fields):
        root_logger.debug("event message: %r", message_fields)

    def _write_hrv_entry(self, fit_file, message_fields):
        root_logger.debug("hrv message: %r", message_fields)

    def _write_ohr_settings_entry(self, fit_file, message_fields):
        root_logger.debug("ohr_settings message: %r", message_fields)

    def _write_software_entry(self, fit_file, message_fields):
        root_logger.debug("software message: %r", message_fields)

    def _write_file_creator_entry(self, fit_file, message_fields):
        root_logger.debug("file creator message: %r", message_fields)

    def _write_sport_entry(self, fit_file, message_fields):
        root_logger.debug("sport message: %r", message_fields)

    def _write_sensor_entry(self, fit_file, message_fields):
        root_logger.debug("sensor message: %r", message_fields)

    def _write_source_entry(self, fit_file, message_fields):
        root_logger.debug("source message: %r", message_fields)

    def _write_training_file_entry(self, fit_file, message_fields):
        root_logger.debug("Training file entry: %r", message_fields)

    def _write_attribute(self, timestamp, message_fields, attribute_name, db_attribute_name=None):
        attribute = message_fields.get(attribute_name)
        if attribute is not None:
            if db_attribute_name is None:
                db_attribute_name = attribute_name
            Attributes.s_set_newer(self.garmin_db_session, db_attribute_name, attribute, timestamp)

    def _write_attributes(self, timestamp, message_fields, attribute_names):
        for attribute_name in attribute_names:
            self._write_attribute(timestamp, message_fields, attribute_name)

    def _write_measurement_ssytem_attributes(self, timestamp, message_fields):
        for attribute_name in ['dist_setting', 'speed_setting', 'height_setting', 'temperature_setting']:
            self._write_attribute(timestamp, message_fields, attribute_name, 'measurement_system')

    def _write_device_settings_entry(self, fit_file, message_fields):
        root_logger.debug("device settings message: %r", message_fields)
        timestamp = fit_file.time_created_local
        self._write_attributes(timestamp, message_fields, ['active_time_zone', 'date_mode'])
        self._write_measurement_ssytem_attributes(timestamp, message_fields)
        self._write_attribute(timestamp, message_fields, 'active_time_zone', 'time_zone')
        self._write_attribute(timestamp, message_fields, 'date_mode', 'date_format')

    def _write_battery_entry(self, fit_file, message_fields):
        root_logger.debug("battery message: %r", message_fields)

    def _write_user_profile_entry(self, fit_file, message_fields):
        root_logger.debug("user profile message: %r", message_fields)
        timestamp = fit_file.time_created_local
        attribute_names = [
            'gender', 'height', 'weight', 'language', 'dist_setting', 'weight_setting', 'position_setting', 'elev_setting', 'sleep_time', 'wake_time',
            'speed_setting'
        ]
        self._write_attributes(timestamp, message_fields, attribute_names)
        self._write_measurement_ssytem_attributes(timestamp, message_fields)

    def _write_activity_entry(self, fit_file, message_fields):
        root_logger.debug("activity message: %r", message_fields)

    def _write_zones_target_entry(self, fit_file, message_fields):
        root_logger.debug("zones target message: %r", message_fields)

    def _write_dev_data_id_entry(self, fit_file, message_fields):
        root_logger.debug("dev_data_id message: %r", message_fields)

    def _write_field_description_entry(self, fit_file, message_fields):
        root_logger.debug("field_description message: %r", message_fields)

    def _write_length_entry(self, fit_file, message_fields):
        root_logger.debug("length message: %r", message_fields)

    def _write_set_entry(self, fit_file, message_fields):
        root_logger.debug("set message: %r", message_fields)

    def _write_watchface_settings_entry(self, fit_file, message_fields):
        root_logger.debug("watchface settings message: %r", message_fields)

    def _write_personal_record_entry(self, fit_file, message_fields):
        logger.info("personal record message: %r", message_fields)



"""Class that specializes TCX files for GarminDb."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import re
from cached_property import cached_property

import tcxfile
from idbutils import Location
from fitfile import Distance, Speed, conversions

from .garmindb import Device


class Tcx(tcxfile.Tcx):
    """Read and write TCX files."""

    __product_to_manufactuer_cache = {}

    __default_device_serial_numbers = {
        (Device.Manufacturer.Microsoft, 'Microsoft Band') : Device.unknown_device_serial_number + 1
    }

    def __init__(self, debug=False):
        """Return and instance of the GaminDbTcx class."""
        super().__init__(debug)

    def add_lap(self, start_dt, end_dt, distance, calories):
        """Add a lap to the TCX file data."""
        return super().add_lap(start_dt, end_dt, distance.to_meters(), calories)

    def add_point(self, track, dt, location, alititude, heart_rate, speed):
        """Add a point to the lap."""
        return super().add_point(track, dt, (location.lat_deg, location.long_deg), alititude.to_meters(), heart_rate, speed.to_mps())

    def add_creator(self, product, serial_number, product_id=None, version=None):
        """Add a creator element."""
        super().add_creator(product, serial_number, product_id, version)

    def __manufacturer_from_product(self, product):
        for manufacturer in Device.Manufacturer:
            if manufacturer.name.lower() in product.lower():
                return manufacturer
        mappings = {
            r'VivoActive|Forerunner|Fenix' : Device.Manufacturer.Garmin,
        }
        for regex, manufacturer in mappings.items():
            if re.search(regex, product, re.IGNORECASE):
                return manufacturer

    def _manufacturer_from_product(self, product):
        if product in self.__product_to_manufactuer_cache:
            return self.__product_to_manufactuer_cache[product]
        manufacturer = self.__manufacturer_from_product(product)
        if manufacturer is not None:
            self.__product_to_manufactuer_cache[product] = manufacturer
        return manufacturer

    def get_manufacturer_and_product(self):
        """Return the product and interperlated manufacturer from the parsed TCX file."""
        product = super().creator_product
        if not product:
            return (None, None)
        return (self._manufacturer_from_product(product), product)

    @cached_property
    def serial_number(self):
        """Return the serial number of the device that recorded the parsed TCX file."""
        serial_number = super().creator_serialnumber
        if not serial_number or serial_number == '0':
            (manufactuer, product) = self.get_manufacturer_and_product()
            if (manufactuer, product) in self.__default_device_serial_numbers:
                serial_number = self.__default_device_serial_numbers[(manufactuer, product)]
            else:
                serial_number = Device.unknown_device_serial_number
        return serial_number

    @cached_property
    def start_loc(self):
        """Return the start location of the activity as a Location instance."""
        return Location(location=super().start_loc)

    @cached_property
    def end_loc(self):
        """Return the end location of the activity as a tuple of Location instances."""
        return Location(location=super().end_loc)

    @cached_property
    def distance(self):
        """Return the total distance recorded for the activity."""
        return Distance.from_meters(super().distance)

    @cached_property
    def speed_max(self):
        """Return the maximum of all speed readings in the TCX file."""
        return Speed.from_mps(super().speed_max)

    @cached_property
    def ascent(self):
        """Return the total ascent over the activity."""
        return Distance.from_meters(super().ascent)

    @cached_property
    def descent(self):
        """Return the total descent over the activity."""
        return Distance.from_meters(super().descent)

    def get_lap_duration(self, lap):
        """Return the recorded duration for the lap."""
        return conversions.secs_to_dt_time(super().get_lap_duration(lap))

    def get_lap_distance(self, lap):
        """Return the recorded distance for the lap."""
        return Distance.from_meters(super().get_lap_distance(lap))

    def get_point_loc(self, point):
        """Return the position of the trackpoint."""
        return Location(location=super().get_point_loc(point))

    def get_point_altitude(self, point):
        """Return the altitude of the trackpoint."""
        return Distance.from_meters(super().get_point_altitude(point))

    def get_point_speed(self, point):
        """Return the speed readings in the point."""
        return Speed.from_mps(super().get_point_speed(point))



"""Test FIT file parsing."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging
import datetime
import re

import fitfile
from idbutils import FileProcessor


root_logger = logging.getLogger()
handler = logging.FileHandler('fit_file.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


test_activity_files     = True
test_monitoring_files   = True
test_sleep_files        = True
test_metrics_files      = True
test_unknown_files      = True


class TestFitFile(unittest.TestCase):
    """Class for testing FIT file parsing."""

    @classmethod
    def setUpClass(cls):
        cls.measurement_system = fitfile.field_enums.DisplayMeasure.statute
        cls.file_path = 'test_files/fit'

    def check_message_fields(self, fit_file, message_type, message):
        unknown_message_fields = {}
        self.check_timestamp(fit_file, message)
        self.check_temperature(message)
        self.check_sport_value(fit_file, message)
        for field_name in message.field_values:
            if not message.field_values[field_name].is_invalid() and field_name.startswith('unknown'):
                if message_type not in unknown_message_fields:
                    logger.info("Unknown %s message field: %s value %s", message_type, field_name, message.fields[field_name])
                    unknown_message_fields[message_type] = [field_name]
                elif field_name not in unknown_message_fields[message_type]:
                    logger.info("Unknown %s message field: %s value: %s", message_type, field_name, message.fields[field_name])
                    unknown_message_fields[message_type].append(field_name)

    def check_message_types(self, fit_file, dump_message=False):
        unknown_messages = []
        for message_type in fit_file.message_types:
            if message_type.name.startswith('unknown'):
                if message_type.name not in unknown_messages:
                    logger.info("Unknown message type: %s in %s", message_type.name, fit_file.type)
                    unknown_messages.append(message_type.name)
            messages = fit_file[message_type]
            for message in messages:
                if dump_message:
                    logger.info("Message: %r", message)
                self.check_message_fields(fit_file, message_type, message)

    def check_type(self, fit_file, message, key, expected_type):
        if key in message.fields:
            value = message.fields[key]
            self.assertIsInstance(value, expected_type, 'file %s expected %r found %r' % (fit_file.filename, expected_type, value))
            logger.info("%s %r: %r", fit_file.filename, message.type, value)

    def check_value(self, fit_file, message, field_name, expected_value):
        if field_name in message.fields:
            value = message.fields[field_name]
            self.assertEqual(value, expected_value, 'file %s expected %r found %r' % (fit_file.filename, expected_value, value))

    def check_value_range(self, fit_file, message, field_name, min_value, max_value, round_value=False):
        if field_name in message.fields:
            value = message.fields[field_name]
            if value is not None:
                if round_value:
                    value = round(value)
                self.assertGreaterEqual(value, min_value, '%s expected greater than %r was %r: %s %r' %
                                        (field_name, min_value, value, fit_file.filename, message))
                self.assertLess(value, max_value, '%s expected less than %r was %r: %s %r ' %
                                (field_name, max_value, value, fit_file.filename, message))
                return True

    def check_timestamp(self, fit_file, message):
        # Garmin Connect generated files can have device dates far in the future
        if (message.type != fitfile.MessageType.device_info or message.type != fitfile.MessageType.file_id) and message.fields.product != fitfile.GarminProduct.connect:
            self.check_value_range(fit_file, message, 'timestamp',
                                   datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc),
                                   datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc))

    def check_temperature(self, message):
        for field_name in message.fields:
            if re.search('temperature_?\a{3}', field_name):
                logger.info("checking " + field_name)
                self.check_value_range(message, field_name, 0, 100)

    def check_timestamp_delta(self, fit_file, start_time, end_time, bounds):
        self.assertEqual(start_time.tzinfo, end_time.tzinfo, 'timezones do not match')
        self.assertGreaterEqual(start_time, self.start_time, 'timestamp before file start')
        self.assertLessEqual(end_time, self.end_time, 'timestamp after end of file')
        total_seconds_in_file = (end_time - start_time).total_seconds()
        self.assertGreaterEqual(total_seconds_in_file, bounds[0], f'time span ({start_time}, {end_time}) is negative')
        self.assertLessEqual(total_seconds_in_file, bounds[1], f'time for {fit_file.filename} span ({start_time}, {end_time}) greater than bound')

    def check_file_id(self, fit_file, file_type):
        self.assertGreaterEqual(fit_file.utc_offset, -36000, 'Is not a valid time zone offset')
        self.assertLessEqual(fit_file.utc_offset, +46800, 'Is not a valid time zone offset')
        # file contains less than a day span of time
        (self.start_time, self.end_time) = fit_file.date_span()
        if fit_file.product != fitfile.GarminProduct.connect:
            self.check_timestamp_delta(fit_file, self.start_time, self.end_time, (0, 86400))
        for message in fit_file.file_id:
            self.check_value(fit_file, message, 'manufacturer', fitfile.Manufacturer.Garmin)
            self.check_value(fit_file, message, 'type', file_type)

    def check_sport_value(self, fit_file, message):
        self.check_type(fit_file, message, 'sport', fitfile.Sport)
        self.check_type(fit_file, message, 'sub_sport', fitfile.SubSport)

    def check_sport(self, fit_file):
        sport = None
        sub_sport = None
        for sport_message in fit_file.sport:
            self.check_sport_value(fit_file, sport_message)
            sport = sport_message.fields.sport
            sub_sport = sport_message.fields.sub_sport
        logger.info("%s: %r %r", fit_file.filename, sport, sub_sport)
        return (sport, sub_sport)

    def check_step_message(self, fit_file, message_index, message):
        self.check_value_range(fit_file, message, 'avg_vertical_oscillation', 0, 10)
        self.check_value_range(fit_file, message, 'step_length', 0, 64)
        self.check_value_range(fit_file, message, 'speed', 0, 25)

    def check_step_record(self, fit_file, message_index, message):
        self.check_value_range(fit_file, message, 'distance', 0, 100)
        self.check_step_message(fit_file, message_index, message)

    def check_step_lap_or_session(self, fit_file, message_index, message):
        self.check_value_range(fit_file, message, 'distance', 0, 100 * 5280)
        self.check_step_message(fit_file, message_index, message)

    def check_record(self, fit_file, sport, sub_sport, message_index, message):
        self.check_message_fields(fit_file, message.type, message)
        if sport == fitfile.Sport.running or sport == fitfile.Sport.walking or sub_sport == fitfile.SubSport.elliptical:
            self.check_step_record(fit_file, message_index, message)

    def check_lap_or_session(self, fit_file, sport, sub_sport, message_index, message):
        self.check_message_fields(fit_file, message.type, message)
        if sport == fitfile.Sport.running or sport == fitfile.Sport.walking or sub_sport == fitfile.SubSport.elliptical:
            self.check_step_lap_or_session(fit_file, message_index, message)

    def check_monitoring_messages(self, fit_file):
        last_steps = {}
        last_timestamp = None
        for message in fit_file.monitoring:
            if 'steps' in message.fields:
                steps = message.fields.steps
                activity_type = message.fields.activity_type
                if activity_type in last_steps:
                    activity_last_steps = last_steps[activity_type]
                    self.assertGreaterEqual(steps, activity_last_steps, f'{fit_file.filename}: {repr(message)} - steps not greater than last steps')
                last_steps[activity_type] = steps
                if last_timestamp:
                    self.check_timestamp_delta(fit_file, last_timestamp, message.fields.timestamp, (0, 43200))
                last_timestamp = message.fields.timestamp
            self.check_message_fields(fit_file, message.type, message)
            self.check_value_range(fit_file, message, 'distance', 0, 100 * 5280, True)
            self.check_value_range(fit_file, message, 'cum_ascent', 0, 5280, True)
            self.check_value_range(fit_file, message, 'cum_descent', 0, 5280, True)

    def check_monitoring_file(self, filename):
        fit_file = fitfile.file.File(filename, self.measurement_system)
        self.check_message_types(fit_file, dump_message=True)
        logger.info('%s (%s) monitoring file message types: %s', filename, fit_file.time_created_local, fit_file.message_types)
        self.check_file_id(fit_file, fitfile.FileType.monitoring_b)
        self.check_monitoring_messages(fit_file)

    def check_activity_file(self, filename):
        fit_file = fitfile.file.File(filename, self.measurement_system)
        logger.info('%s (%s) activity file message types: %s', filename, fit_file.time_created_local, fit_file.message_types)
        self.check_message_types(fit_file, dump_message=True)
        self.check_file_id(fit_file, fitfile.FileType.activity)
        (sport, sub_sport) = self.check_sport(fit_file)
        for message_index, message in enumerate(fit_file.record):
            self.check_record(fit_file, sport, sub_sport, message_index, message)
        for message_index, message in enumerate(fit_file.lap):
            self.check_lap_or_session(fit_file, sport, sub_sport, message_index, message)
        for message_index, message in enumerate(fit_file.session):
            self.check_lap_or_session(fit_file, sport, sub_sport, message_index, message)

    def check_sleep_file(self, filename):
        fit_file = fitfile.file.File(filename, self.measurement_system)
        logger.info('%s (%s) sleep file message types: %s', filename, fit_file.time_created_local, fit_file.message_types)
        self.check_message_types(fit_file, dump_message=True)
        self.check_file_id(fit_file, fitfile.FileType.sleep)

    def check_unknown_file(self, filename):
        logger.info('Parsing ' + filename)
        fit_file = fitfile.file.File(filename, self.measurement_system)
        logger.info('%s (%s) unknown file message types: %s', filename, fit_file.time_created_local, fit_file.message_types)
        self.check_message_types(fit_file, dump_message=True)

    #
    # The tests
    #
    @unittest.skipIf(not test_monitoring_files, 'Test not selected')
    def test_parse_monitoring(self):
        monitoring_path = self.file_path + '/monitoring'
        file_names = FileProcessor.dir_to_files(monitoring_path, fitfile.file.name_regex, False)
        for file_name in file_names:
            self.check_monitoring_file(file_name)

    @unittest.skipIf(not test_activity_files, 'Test not selected')
    def test_parse_activity(self):
        activity_path = self.file_path + '/activity'
        file_names = FileProcessor.dir_to_files(activity_path, fitfile.file.name_regex, False)
        for file_name in file_names:
            self.check_activity_file(file_name)

    @unittest.skipIf(not test_sleep_files, 'Test not selected')
    def test_parse_sleep(self):
        activity_path = self.file_path + '/sleep'
        file_names = FileProcessor.dir_to_files(activity_path, fitfile.file.name_regex, False)
        for file_name in file_names:
            self.check_sleep_file(file_name)

    @unittest.skipIf(not test_metrics_files, 'Test not selected')
    def test_parse_metrics(self):
        # root_logger.setLevel(logging.DEBUG)
        activity_path = self.file_path + '/metrics'
        file_names = FileProcessor.dir_to_files(activity_path, fitfile.file.name_regex, False)
        for file_name in file_names:
            self.check_unknown_file(file_name)

    @unittest.skipIf(not test_unknown_files, 'Test not selected')
    def test_parse_unknown(self):
        # root_logger.setLevel(logging.DEBUG)
        activity_path = self.file_path + '/unknown'
        file_names = FileProcessor.dir_to_files(activity_path, fitfile.file.name_regex, False)
        for file_name in file_names:
            self.check_unknown_file(file_name)


if __name__ == '__main__':
    unittest.main(verbosity=2)
